/*   File Header   ---------------------------------
     File Name:        db/message.c
     Author and Date:  Paul B. Anderson, June 1990
     Description:      Format and output error and notification
                       messages.
     End Header   ---------------------------------- */


#include <stdio.h>
#include <geodata_private.h>

/*   Function Header   -----------------------------
     Description:      Display a message to the client
     Parameters:       class:   An integer having one of the
                                following values:
                         GB_ERROR  -- Format as an error message
                                      and write to stderr
                         GB_EPRINT -- Write text string to stderr
                                      without formatting
                         GB_NOTIFY -- Format as a notification message
                                      and write to stdout
                         GB_NPRINT -- Write text string to stdout
                                      without formatting
                       place    A character string containing the name
                                of the function originating the message
                       text     A character string containing the message
                                text.  Any variable parameters should
                                have been previously incorporated using
                                sprintf or equivalent.
     Preconditions:    None
     Postconditions:   No change of state
     Return Value:     None
     Logic:
       1.   Use class parameter to select proper action.
       2.   case GB_ERROR:   Format as an error message (incorporating
                             the place and text strings.  Write to stderr.
       3.   case GB_EPRINT:  Write text string to stderr.
       4.   case GB_NOTIFY:  Format as a notification message using only
                             the text string.  Write to stdout.
       5.   case GB_NPRINT:  Write text string to stdout.
     End Header   ---------------------------------- */

void GB_message(class,place,text)
  int class;  char *place, *text;
{
  switch(class) {
  case GB_ERROR:
    fprintf(stderr,"\n;***GeoData Error ***  In routine %s,\n;*** %s\n",place,text);
    break;
  case GB_EPRINT:
    fprintf(stderr,"%s\n",text);
    break;
  case GB_NOTIFY:
    fprintf(stderr,"\n;===GeoData Note ===  In routine %s,\n;=== %s\n",place,text);
    break;
  case GB_NPRINT:
    fprintf(stdout,"%s\n",text);
    break;
  }
}
