#include <stdio.h>
#include <geodata.h>
#include <geodata_private.h>
#include "types.h"

#ifndef GEODATA_REMOTE_DB 
#include <cm/paris.h>
#include <cm/cmfs.h>
#include <cm/cm_file.h>
#include <cm/cm_stat.h>
/* should be an absolute pathname here*/
#include <pgl/pgl.h>
#include <pgl/pgl_internal.h>
#endif

static unsigned int video_rec,video_play,play_no,rec_no;

void GB_zero_float_vector(x,n)
     float x[];
     int n;
{
  int i;
  for (i=0;i<n;i++)
    x[i]=0.0;
}

void GB_write_scene();
void GB_write_table();
void GB_write_rep_list();
void GB_write_representation();
void GB_write_data_desc();
void GB_write_polygon();
void GB_write_vertex();
void GB_write_symbol();
void GB_write_display();
void GB_write_xtool_projection();
void GB_write_line_style();
void GB_write_movie();
void GB_write_video();

void GB_recur_write_scene();
void GB_recur_write_table();
void GB_recur_write_rep_list();
void GB_recur_write_representation();
void GB_recur_write_data_desc();
void GB_recur_write_polygon();
void GB_recur_write_vertex();
void GB_recur_write_symbol();
void GB_recur_write_display();
void GB_recur_write_xtool_projection();
void GB_recur_write_line_style();
void GB_recur_write_movie();
void GB_recur_write_video();

void GB_write_instance();

char *GB_create_scene()
{
  GS_scene_t new =(GS_scene_t)
    malloc(sizeof(GS_scene));

  new->projection = GS_MILLER;
  GB_zero_float_vector(new->parameters,NO_PROJECTION_PARAMS);
  GB_zero_float_vector(new->center,3);
  GB_zero_float_vector(new->light_source,3);
  GB_zero_float_vector(new->slice,4);
  (new->dimensions)[0]=512;
  (new->dimensions)[1]=512;
  new->zoom=0;
  new->representations = (GS_rep_list_t)0;
  new->using_slice=0;
  new->light_type = GS_AMBIENT;
  new->ambient=1.0;
  new->buffer=0;
  new->remote_render_t=0;
  new->background[0]=0;
  new->background[1]=0;
  new->background[2]=0;
  new->time_this_scene=0;
  new->record_this_scene=0;
  new->depth_bits=10;
  new->priority_bits=3;
  return((char *)new);  
}


void GB_delete_scene(instance)
     GS_scene_t instance;
{
/*  if (instance->buffer !=0)
    PGL_deallocate_buffer(instance->buffer);*/
  free(instance);
}



char *GB_create_table()
{
  GS_table_t new =(GS_table_t)malloc(sizeof(GS_table));
  new->next=0;
  new->discrete=0;
  new->scale=new->rotation=0.0;
  new->color[0]=new->color[1]=new->color[2]=0;
  new->value=0.0;
  new->opacity=1.0;
  return((char *)new);  
}


char *GB_next_table(instance)
     char *instance;
{
  return((char *)(&((GS_table_t)instance)->next));
}


void GB_delete_table(instance)
     GS_table_t instance;
{
  free(instance);
}


char *GB_create_rep_list()
{
  GS_rep_list_t new =(GS_rep_list_t)
    malloc(sizeof(GS_rep_list));
  new->next=0;
  new->rep_id=0;
  new->data_id=0;
  new->priority=0;
  new->rendering_memory=0;
  return((char *)new);  
}

void GB_delete_rep_list(instance)
     GS_rep_list_t instance;
{
  free(instance);
}


char *GB_next_rep_list(instance)
     char *instance;
{
  return((char *)(&((GS_rep_list_t)instance)->next));
}



char *GB_create_polygon()
{
  GS_polygon_t new =(GS_polygon_t)
    malloc(sizeof(GS_polygon));
  new->vertices = (GS_vertex_t)0;
  new->next = (GS_polygon_t)0;
  return((char *)new);  
}

/* free the verticies*/
void GB_delete_polygon(instance)
     GS_polygon_t instance;
{
  free(instance);
}


char *GB_next_polygon(instance)
     char *instance;
{
  return((char *)(&((GS_polygon_t)instance)->next));
}




char *GB_create_vertex()
{
  GS_vertex_t new =(GS_vertex_t)
    malloc(sizeof(GS_vertex));
  new->next = (GS_vertex_t)0;
  return((char *)new);  
}

/* free the verticies*/
void GB_delete_vertex(instance)
     GS_vertex_t instance;
{
  free(instance);
}


char *GB_next_vertex(instance)
     char *instance;
{
  return((char *)(&((GS_vertex_t)instance)->next));
}


char *GB_create_symbol()
{
  GS_symbol_t new =(GS_symbol_t)
    malloc(sizeof(GS_symbol));
  new->polygons = (GS_polygon_t)0;
  return((char *)new);  
}

/* free the verticies*/
void GB_delete_symbol(instance)
     GS_symbol_t instance;
{
  free(instance);
}


char *GB_create_rep()
{
  GS_representation_t new = (GS_representation_t)
    malloc(sizeof(GS_representation));  
  new->tables =(GS_table_t)GB_create_table();
  return((char *)new);  
}

void GB_delete_rep(instance)
     GS_representation_t instance;
{
  GS_table_t k, temp;
  for (k=instance->tables;k!=(GS_table_t)0;k=temp){ 
    temp = k->next;
    free(k);
  }
  free(instance);
}


char *GB_create_data_desc()
{
  GD_data_desc *new;  
  if ((new = (GD_data_desc *) malloc(sizeof(GD_data_desc)))
      == (GD_data_desc *)0) return((char *)0);
  bzero(new->detail,DETAIL_LENGTH);
  bzero(new->filename,DETAIL_LENGTH);
  new->coord_syst = 0;
  GB_zero_float_vector(new->units,3);
  new->type = 0;
  new->rank = 0;
  new->fortran_data_flag = 0; /* NOT fortran cm data */
  new->fe_shape[0]=0;
  new->fe_shape[1]=0;
  new->fe_shape[2]=0;
  new->cm_fe_flag = 0;
  new->fe_data = 0;
  new->context = 0;
  new->values = 0;
  new->transfer = 0;
  new->user_yp = 0;
  new->surface = 0.0;
  new->proj_fn = 0;
  new->exaggeration=1.0;
  new->transform_t=1;
  new->load_type = GD_CLIENT;
  new->periodic_axes[0]=GD_NON_PERIODIC;
  new->periodic_axes[1]=GD_NON_PERIODIC;
  new->periodic_axes[2]=GD_NON_PERIODIC;
  GB_zero_float_vector(new->r,3);
  GB_zero_float_vector(new->nv,3);
  GB_zero_float_vector(new->yp,3);
  GB_zero_float_vector(new->origin,3);
  GB_zero_float_vector(new->extent,3);
  new->surface = GD_MSL;
  new->span[0]=new->span[1]=new->span[2]=1;
  new->vp_set=(CM_vp_set_id_t *)malloc(sizeof(CM_vp_set_id_t));
  *(new->vp_set)=(-1);
  GB_zero_float_vector(new->proj_params,10);
  new->long_lines_flag = 0;
  new->print_stats=0;
  return((char *)new);
} 


void GB_delete_data_desc(instance)
     GD_data_desc *instance;
{
  if (instance == 0) return;
    /* Delete context, values, and transfer fields here */
/*  if (instance->ref_resolution != 0){

    GB_delete_ref_resolution_list(instance->ref_resolution);
  }*/
  free (instance);
}


char *GB_create_display()
{
  GS_display_t new =(GS_display_t)malloc(sizeof(GS_display));
  char emsg[64];
  
  new->display_type = GS_CMFB;
  new->size[0]=512;
  new->size[1]=512;
  new->name[0]=0;
  return((char *)new);  
}

void GB_execute_display (display)
     GS_display_t display;
{
  int zoom;
  char emsg[64];

#ifndef GEODATA_REMOTE_DB
  switch (display->display_type){
  case GS_CMFB:
    display->display_id =(char *)
      PGL_allocate_display(PGL_cmfb,(display->name[0]==0)?
			   ((char*)getenv("CM_FRAMEBUFFER")):
			   display->name,
			   display->size[0],
			   display->size[1]);
    break;
  case GS_XWINDOWS:
    display->display_id =(char *)
      PGL_allocate_display(PGL_xwindows,display->name,display->size[0],
			   display->size[1]);
    break;
  case GS_CMFS:    
    display->display_id =(char *)
      PGL_allocate_display(PGL_cmfs,display->name,display->size[0],
			   display->size[1]);
    break;
  case GS_GLWINDOWS:
    display->display_id =(char *)
      PGL_allocate_display(PGL_sgi,display->name,display->size[0],
			   display->size[1]);
    break;
  }
#endif
}


void GB_delete_display(instance)
     GS_display_t instance;
{
#ifndef GEODATA_REMOTE_DB    
  PGL_deallocate_display(instance->display_id);
#endif
  free(instance);
}

char *GB_create_xtool_projection()
{
  int i;
  GX_xtool_projection_t new =
    (GX_xtool_projection_t)malloc(sizeof(GX_xtool_projection));
  new->GX_proj_code = GS_NO_PROJECTION;
  for (i=0; i<NO_PROJECTION_PARAMS; i++) {
    new->GX_defaults[i] = 0;
    new->GX_convert[i] = 1.0;
    new->GX_min[i] = 0;
    new->GX_max[i] = 1;
    sprintf (new->GX_labels[i], "(Unused)");
  }
  return((char *)new);  
}

void GB_delete_xtool_projection(instance)
     GX_xtool_projection_t instance;
{
  free(instance);
}

char *GB_create_update_scene()
{
  GB_update_scene_t new = (GB_update_scene_t)
    malloc(sizeof(GB_update_scene));  
  return((char *)new);  
}

void GB_execute_update_scene(instance)
     GB_update_scene_t instance;
{
#ifndef GEODATA_REMOTE_DB    
  GS_update_scene (instance->scene_id);
#endif
/*  free(instance);*/
}


char *GB_create_display_scene()
{
  GB_display_scene_t new = (GB_display_scene_t)
    malloc(sizeof(GB_display_scene));  
  (new->offset[0])=((new->offset[1])=0);
  return((char *)new);  
}


void GB_execute_display_scene(instance)
     GB_display_scene_t instance;
{
#ifndef GEODATA_REMOTE_DB    
  GS_display_scene (instance->display_id,instance->scene_id,
		    (instance->offset[0]),(instance->offset[1]));
#endif
/*  free(instance);*/
}


char *GB_create_include()
{
  GB_include_t new = (GB_include_t)malloc(sizeof(GB_include));  
  return((char *)new);  
}


void GB_execute_include(instance)
     GB_include_t instance;
{
  GB_load_object_file (instance->file);
/*  free(instance);*/
}

GB_entry GB_hash_table[];

char *GB_create_print_objects()
{
  GB_print_objects_t new = (GB_print_objects_t)malloc(sizeof(GB_print_objects));  
  return((char *)new);  
}

void GB_execute_print_objects(instance,frame)
     GB_print_objects_t instance;
     GB_parse_frame_t frame;
{
  int i,k;
  GB_entry *j;
  char type_name[GB_LABEL_LENGTH];
  
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    for (j = GB_hash_table+i; j != (GB_entry *)0;j=j->next)
      if (j->id){
	GB_encode_constant (type_name,j->type);
	for (k=0;type_name[k]!=0;k++)
	  if ((type_name[k] >= 'A') && (type_name[k] <= 'Z'))
	    type_name[k]-='A'-'a';
	fprintf(frame->out, "(%s %s)\n",type_name,j->label);
      }
  fprintf(frame->out, ";end\n");
}


char *GB_create_print()
{
  GB_print_t new = (GB_print_t)malloc(sizeof(GB_print));  
  return((char *)new);  
}

void GB_execute_print(instance,frame)
     GB_print_t instance;
     GB_parse_frame_t frame;
{
  GB_describe_object (instance->id,frame->out,0);
}


char *GB_create_recur_print()
{
  GB_recur_print_t new = (GB_recur_print_t)malloc(sizeof(GB_recur_print));  
  new->out_file[0]='\0';
  return((char *)new);  
}

void GB_execute_recur_print(instance,frame)
     GB_recur_print_t instance;
     GB_parse_frame_t frame;
{
  FILE *fp;
  if(instance->out_file[0])fp=fopen(instance->out_file,"w");
  else fp=stdout;
  if(fp) GB_recur_describe_object (instance->obj_id,fp,0);
  else GB_message (GB_ERROR, "recur_printor", "Bad file name");
  if(fp!=stdout)fclose(fp);
}


char *GB_create_map_data()
{
  GB_map_data_t new = (GB_map_data_t)malloc(sizeof(GB_map_data));  
  return((char *)new);  
}


void GB_execute_map_data(instance,frame)
     GB_map_data_t instance;
     GB_parse_frame_t frame;
{
#ifndef GEODATA_REMOTE_DB    
  GD_map (instance->to,instance->from);
#endif
}


char *GB_create_movie()
{
  GS_movie_t new = (GS_movie_t)malloc(sizeof(GS_movie));
  new->frames_per_second=30;
  new->frame_offset=0;
  new->frames=0;
  return((char *)new);  
}


char *GB_create_play_movie()
{
  GB_play_movie_t new = (GB_play_movie_t)malloc(sizeof(GB_play_movie));  
  return((char *)new);  
}


void GB_execute_play_movie(instance,frame)
     GB_play_movie_t instance;
     GB_parse_frame_t frame;
{
#ifndef GEODATA_REMOTE_DB 
  int fd,count,i;
  struct cm_stat sbuff;
  CM_field_id_t field;
  char *my_display;
  PGL_buffer_t temp_buffer;
  unsigned dimensions[2];
  CM_geometry_id_t fg;
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  GS_movie_t movie=(GS_movie_t)GB_get_object(instance->movie_id);
  char *pathname;
  
  my_display=((GS_display_t)GB_get_object(instance->display_id))->display_id;
  
  if (fd=CMFS_open(movie->movie_filename,CMFS_O_RDONLY) < 0){
    CMFS_perror ("open failed");
    exit(1);
  }
  fg=(CM_geometry_id_t)CMFS_file_geometry(fd);
  dimensions[0]=CM_geometry_axis_length(fg,0);
  dimensions[1]=CM_geometry_axis_length(fg,1);
  CM_set_vp_set(CM_allocate_vp_set(CM_create_geometry(dimensions,2)));

  CMFS_fstat(fd,&sbuff);
  count = sbuff.cmst_size/(dimensions[0]*dimensions[1]*3);
  field = CM_allocate_stack_field(24*count);

  /*  CMFS_fcntl(fd,CMFS_F_SETFL,CMFS_FSTREAMING);*/
  /*  CMFS_read_file_always(fd,0,count*24);*/
  /*  CMFS_streaming_iostat(fd);*/
  /*  CMFS_partial_read_file_always(fd,field,24);*/
  /*  CMFS_streaming_iostat(fd);*/
  
  CMFS_read_file_always(fd,field,24*count);
  CMFS_close(fd);

  for(i=movie->frame_offset;
      i<MIN((movie->frame_offset+movie->frames),count);
      i++){
    temp_buffer=PGL_wrap_buffer_around_field(field+24*i,
					     PGL_overwrite,
					     "ru8 gu8 bu8");
    PGL_display_buffer(my_display,temp_buffer,
		       instance->display_offset[0],
		       instance->display_offset[1]);
    PGL_deallocate_buffer(temp_buffer);
  }
  CM_set_vp_set(entry_vp_set);
#endif
}


char *GB_create_video()
{

 GS_video_t new = (GS_video_t)malloc(sizeof(GS_video));
  new->disk=0;
  new->speed=30;
  new->start_frame=0;
  new->stop_frame=1;
  new->description[0]='\0';
  return((char *)new);  
}

char *GB_create_open_video_display()
{

  GB_open_video_display_t new = 
    (GB_open_video_display_t)malloc(sizeof(GB_open_video_display));
  return((char *)new);  
}

char *GB_create_close_video_display()
{

  GB_close_video_display_t new = 
    (GB_close_video_display_t)malloc(sizeof(GB_close_video_display));
  return((char *)new);  
}

char *GB_create_play_video()
{

  GB_play_video_t new = (GB_play_video_t)malloc(sizeof(GB_play_video));  
  return((char *)new);  
}

char *GB_create_seek_to_frame()
{

  GB_seek_to_frame_t new =(GB_seek_to_frame_t)malloc(sizeof(GB_seek_to_frame));
  new->seek_pana=0;
  new->frame=(-1);
  new->plus=0;
  new->minus=0;
  return((char *)new);  
}

char *GB_create_set_video_switch()
{

  GB_set_video_switch_t new =
    (GB_set_video_switch_t)malloc(sizeof(GB_set_video_switch));
  new->set_to=0;
  return((char *)new);  
}

char *GB_create_play_pana()
{

  GB_play_pana_t new =
    (GB_play_pana_t)malloc(sizeof(GB_play_pana));
  new->pana_play=0;
  new->play_from=0;
  new->play_to=0;
  new->speed=30;
  return((char *)new);  
}

void GB_execute_open_video_display(instance,frame)
     GB_open_video_display_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC
  VD_open_panasonics();
  VD_set_video(0);
  rec_no = VD_get_disk_number(0);
  play_no = VD_get_disk_number(1);
#endif
}

void GB_execute_close_video_display(instance,frame)
     GB_close_video_display_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC
  VD_close_panasonics();
#endif
}

void GB_execute_play_video(instance,frame)
     GB_play_video_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC

  char result[16],com[16];
  int pana = -1;
  GS_video_t video=(GS_video_t)GB_get_object(instance->video_id);

/* compare to video->disk */
  if(play_no==video->disk) pana=1;
  if(rec_no==video->disk) pana=0;

  if(pana!=-1){
    VD_set_video(pana);
    VD_play(pana,
	    video->start_frame,
	    video->stop_frame,
	    video->speed);
  }
#endif
}

void GB_execute_seek_to_frame(instance,frame)
     GB_seek_to_frame_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC
#ifndef GEODATA_REMOTE_DB 
  char res[16],com[16];
  int new_frame=0;
  int pana = -1;

  if(instance->seek_pana==GB_PLAYER)pana=1;
  if(instance->seek_pana==GB_RECORDER)pana=0;

  if(instance->frame != -1) VD_set_frame_number(pana,new_frame,0);
  else { 
    if(instance->plus)VD_set_frame_number(pana,0,instance->plus);
    if(instance->minus)VD_set_frame_number(pana,0,-instance->minus);
  }
  instance->frame = -1;
  instance->plus=0;
  instance->minus=0;
#endif
#endif
}

void GB_execute_set_video_switch(instance,frame)
     GB_set_video_switch_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC
#ifndef GEODATA_REMOTE_DB 
  VD_set_video(instance->set_to - GB_RECORDER);
#endif
#endif
}

void GB_execute_play_pana(instance,frame)
     GB_play_pana_t instance;
     GB_parse_frame_t frame;
{
#ifdef GEODATA_HAVE_LIBPANASONIC
#ifndef GEODATA_REMOTE_DB 
  VD_play ((instance->pana_play - GB_RECORDER),
	   instance->play_from,
	   instance->play_to,
	   instance->speed);
  instance->play_from=0;
  instance->play_to=0;
  instance->speed=30;
#endif
#endif
}


char *GB_create_line_style()
{
  GS_line_style_t new = (GS_line_style_t)malloc(sizeof(GS_line_style));  
  new->width=0.0;
  return((char *)new);  
}

void GB_delete_line_style(instance)
     GS_line_style_t instance;
{
  free(instance);
}

char *GB_create_handshake()
{
  GB_handshake_t new = (GB_handshake_t)malloc(sizeof(GB_handshake));  
  return((char *)new);
}


char *GB_create_sleep()
{
  GB_sleep_t new = (GB_sleep_t)malloc(sizeof(GB_sleep));  
  return((char *)new);
}

void GB_execute_sleep(instance,frame)
     GB_sleep_t instance;
     GB_parse_frame_t frame;
{
  sleep(instance->seconds);
}


char *GB_create_die()
{
  GB_die_t new = (GB_die_t)malloc(sizeof(GB_die));  
  return((char *)new);
}


void GB_execute_die(instance,frame)
     GB_die_t instance;
     GB_parse_frame_t frame;
{
  exit(0);
}


char *GB_create_detach()
{
  GB_die_t new = (GB_die_t)malloc(sizeof(GB_die));  
  return((char *)new);
}


void GB_execute_detach(instance,frame)
     GB_die_t instance;
     GB_parse_frame_t frame;
{
#ifndef GEODATA_REMOTE_DB    
  CM_detach();
#endif
}


struct GB_object {
  int type;
  char *(*creator)();
  void (*destructor)();
  char *(*next)();
  void (*printor)();
  void (*executor)();
  void (*recur_printor)();
};


struct GB_object GB_objects[]={
  {GB_SCENE,GB_create_scene,GB_delete_scene,0,GB_write_scene,0,
     GB_recur_write_scene},
  {GB_TABLE,GB_create_table,GB_delete_table,GB_next_table,GB_write_table,0,
     GB_recur_write_table},
  {GB_REP_LIST,GB_create_rep_list,GB_delete_rep_list,
               GB_next_rep_list,GB_write_rep_list,0,GB_recur_write_rep_list},
  {GB_REPRESENTATION,GB_create_rep,GB_delete_rep,0,GB_write_representation,0,
     GB_recur_write_representation}, 
  {GB_DATA_DESC,GB_create_data_desc,GB_delete_data_desc,0,
     GB_write_data_desc,0,
     GB_recur_write_data_desc},
  {GB_POLYGON,GB_create_polygon,GB_delete_polygon,GB_next_polygon,
     GB_write_polygon,0,
     GB_recur_write_polygon},
  {GB_VERTEX,GB_create_vertex,GB_delete_vertex,GB_next_vertex,
     GB_write_vertex,0,
     GB_recur_write_vertex},
  {GB_SYMBOL,GB_create_symbol,GB_delete_symbol,0,GB_write_symbol,0,
     GB_recur_write_symbol},
  {GB_LINE_STYLE,GB_create_line_style,GB_delete_line_style,0,
     GB_write_line_style,0,
     GB_recur_write_line_style},
  {GB_DISPLAY,GB_create_display,GB_delete_display,0,GB_write_display,
     GB_execute_display,GB_recur_write_display},
  {GB_XTOOL_PROJECTION,GB_create_xtool_projection,GB_delete_xtool_projection,
     0,GB_write_xtool_projection,0,GB_recur_write_xtool_projection},
  {GB_DISPLAY_SCENE,GB_create_display_scene,0,0,0,GB_execute_display_scene,0},
  {GB_UPDATE_SCENE,GB_create_update_scene,0,0,0,GB_execute_update_scene,0},
  {GB_INCLUDE,GB_create_include,0,0,0,GB_execute_include,0},
  {GB_PRINT_OBJECTS,GB_create_print_objects,0,0,0,GB_execute_print_objects,0},
  {GB_PRINT,GB_create_print,0,0,0,GB_execute_print,0},
  {GB_RECUR_PRINT,GB_create_recur_print,0,0,0,GB_execute_recur_print,0},
  {GB_MAP_DATA,GB_create_map_data,0,0,0,GB_execute_map_data,0},
  {GB_MOVIE,GB_create_movie,0,0,GB_write_movie,0,GB_recur_write_movie},
  {GB_PLAY_MOVIE,GB_create_play_movie,0,0,0,GB_execute_play_movie,0},
  {GB_HANDSHAKE,GB_create_handshake,0,0,0,0,0},
  {GB_SLEEP,GB_create_sleep,0,0,0,GB_execute_sleep,0},
  {GB_DIE,GB_create_die,0,0,0,GB_execute_die,0},
  {GB_DETACH,GB_create_die,0,0,0,GB_execute_detach,0},
  {GB_VIDEO,GB_create_video,0,0,GB_write_video,0,GB_recur_write_video},
  {GB_PLAY_VIDEO,GB_create_play_video,0,0,0,GB_execute_play_video,0},
  {GB_OPEN_VIDEO_DISPLAY,GB_create_open_video_display,0,0,0,
     GB_execute_open_video_display,0},
  {GB_CLOSE_VIDEO_DISPLAY,GB_create_close_video_display,0,0,0,
     GB_execute_close_video_display,0},
  {GB_SEEK_TO_FRAME,GB_create_seek_to_frame,0,0,0,GB_execute_seek_to_frame,0},
  {GB_SET_VIDEO_SWITCH,GB_create_set_video_switch,0,0,0,
     GB_execute_set_video_switch,0},
  {GB_PLAY_PANA,GB_create_play_pana,0,0,0,GB_execute_play_pana,0},
  {0,0,0,0,0}};


char *GB_create_instance(type)
     int type;
{
  int i;
  for (i=0;GB_objects[i].type!=0;i++)
    if ((GB_objects[i].type == type) && GB_objects[i].creator)
      return((*(GB_objects[i].creator))());
  printf ("don't have a creator for type %d.\n",type);
  *(char *)0 = 1;
}


char *GB_next_instance(instance,type)
     char *instance;
     int type;
{
  int i;
  for (i=0;GB_objects[i].type!=0;i++)
    if ((GB_objects[i].type == type) && GB_objects[i].next)
      return((*(GB_objects[i].next))(instance));
  printf ("don't have a next for type %d.\n",type);
}



void GB_delete_instance(pointer,type)
     char *pointer;
     int type;
{
  int i;
  for (i=0;GB_objects[i].type!=0;i++)
    if ((GB_objects[i].type == type) && GB_objects[i].destructor){
      (*(GB_objects[i].destructor))(pointer);
      return;
    }  
  printf ("don't have a destructor for type %d.\n",type);
}


void GB_write_instance(f,instance,type,label,indent)
     FILE *f;
     char *instance;
     int type;
     char *label;
     int indent;
{
  int i,flag=0;
  for (i=0;GB_objects[i].type!=0;i++)
    if ((GB_objects[i].type == type) && GB_objects[i].printor){
      (*(GB_objects[i].printor))(f,instance,label,indent);
      flag =1;
    }
  if (!flag) printf ("don't have a printor for type %d.\n",type);
}


void GB_execute_instance(pointer,type,frame)
     char *pointer;
     int type;
     GB_parse_frame_t frame;
{
  int i;
  
  for (i=0;GB_objects[i].type!=0;i++){
    if ((GB_objects[i].type == type) && (GB_objects[i].executor)){
      (*(GB_objects[i].executor))(pointer,frame);
      return;
    }
  }  
}

void GB_recur_write_instance(f,instance,type,label,indent)
     FILE *f;
     char *instance;
     int type;
     char *label;
     int indent;
{
  int i,flag=0;
  for (i=0;GB_objects[i].type!=0;i++)
    if ((GB_objects[i].type == type) && GB_objects[i].recur_printor){
      (*(GB_objects[i].recur_printor))(f,instance,label,indent);
      flag =1;
    }
  if (!flag) GB_write_instance(f,instance,type,label,indent);
}

