/******************************************************************************
*                                                                             *
*  BEAM -   Calculates using a frequency domain beam former the beam over     *
*           the set of phones whose complex spectrum is given in CM slice     *
*           real_part and whose wweights are given in the slice real_table.   *
*                                                                             *
*     real_beam     - Slice for the complex beam output                       *
*     real_part     - Slice containing spectrum of the input phones           *
*     real_table    - Slice containing phone weights                          *
*     number_phones - Number of phones                                        *
*                                                                             *
******************************************************************************/

#include <cm/paris.h>


  void BEAM(real_beam, real_part, real_table,number_phones)
  unsigned real_beam, real_part, real_table;
  unsigned number_phones;
{
  unsigned  i ;
  unsigned rb,ib,rp,ip,rt,it;

  rb=real_beam;
  ib=rb+32;
  CM_u_move_zero_1L (rb,64);
  for ( i=0 ; i < number_phones ; i++) {
    rp=real_part+i*64;
    ip=rp+32;
    rt=real_table+i*64;
    it=rt+32;
    CM_f_mult_add_1L (rb,rp,rt,rb,23,8);
    CM_f_mult_subf_1L (rb,ip,it,rb,23,8);
    CM_f_mult_add_1L (ib,rp,it,ib,23,8);
    CM_f_mult_add_1L (ib,ip,rt,ib,23,8);
  }
}
