/******************************************************************************
*                                                                             *
*  BITREV2D - Perfroms the bit reverse scrambling associated with the FFT.    *
*             This will work for either the case when the FFT is across the   *
*             whole CM or when multiple FFTs are performed simultaneously.    *
*             The routine can also shift the data ( representing a time delay *
*             if the data is time series or a phase shift if the data is      *
*             in the frequency domain).  The method used is to set up a loca- *
*             tion in each PE which will contain the PE number, subtract the  *
*             shift desired from the contents of this location, bit reverse   *
*             the bottom p ( where 2^p is the FFT size) bits of the contents  *
*             of this location, and then use the contents as the detination   *
*             PE of a router request.                                         *
*                                                                             *
*    real_in    - pointer to the slice of data on the CM to be scrambled      *
*    p         - log to the base 2 of the FFT size                            *
*    num       - number of bits per PE in the data slice                      *
*    offset    - number of points to shift the data - a right shift           *
*                                                                             *
******************************************************************************/

#include <cm/fast-graph.h>
#include <cm/paris.h>
#include <math.h>

void BITREV2D (real_in,  p, num, offset)
unsigned real_in, p, num;
int offset;
{
  unsigned i, pcal, name, save_context_flag;
  unsigned n = 1 << p;
  CM_geometry_id_t current_geometry;
  unsigned numb, temp;
  unsigned sa,x,y;

  current_geometry = CM_vp_set_geometry(CM_current_vp_set);
  pcal = CM_geometry_send_address_length(current_geometry);
  save_context_flag = CM_allocate_stack_field(1);
  CM_store_context(save_context_flag);
  CM_set_context();
  name = CM_allocate_stack_field(pcal);
  CM_u_move_zero_1L(name, pcal);
  CM_my_send_address_1L(name);


  numb=CM_allocate_stack_field(pcal);
  temp=CM_allocate_stack_field(num);

  sa=CM_allocate_stack_field(pcal);
  x=CM_allocate_stack_field(p);
  y=CM_allocate_stack_field(pcal-p);
  CM_my_send_address_1L (sa,pcal);
  CM_extract_news_coordinate_1L(current_geometry,x,0,sa,p);
  CM_u_subtract_constant_2_1L ( x,offset,p);
  CM_move_reversed_1L (x, x, p); 
  CM_extract_news_coordinate_1L(current_geometry,y,1,sa,pcal-p);
  CM_make_news_coordinate_1L ( current_geometry,sa,0,x,p);
  CM_deposit_news_coordinate_1L (current_geometry,sa,1,y,pcal-p);
  CM_send_1L ( temp,sa,real_in,num,CM_no_field);  

  CM_u_move_1L (real_in,temp,num);

  CM_load_context(save_context_flag);
  CM_deallocate_stack_through(save_context_flag);
}
