/*

   CMATMULT - Multiplies the complex matrix stored in a slice in CM memory 
              by the complex matrix in a second slice in PE memory. Call-
              ing sequence is

          CMATMULT ( out, in1,in2,m,p,n)

  where

          out    - Pointer to the slice in CM memory  where 
                   the result is to be stored
          in1    - Pointer to first input matrix
          in2    - Pointer to second input matrix
          m      - Number of rows of first input matrix
          p      - Number of columns of first input matrix =
                   Number of rows of seconfd input matrix
          n      - Number of columns of second input matrix

*/

#include <cm/paris.h>


  void CMATMULT(out, in1, in2, m, p, n)
  unsigned in1,in2,out;
  int m,p,n;
{

  unsigned temp;
  int i,j,k;

  temp=CM_allocate_stack_field(64);
  
  for (i=0;i<m;i++) {
    for (j=0;j<n;j++) {
      CM_c_move_zero_1L (out+64*i+64*m*j,23,8);
      for (k=0;k<p;k++) {
        CMULT(temp,in1+64*i+64*m*k,in2+64*k+64*p*j);
        CM_c_add_2_1L (out+64*i+64*m*j,temp,23,8);
       }
     }
  }

  CM_deallocate_stack_through(temp);

}
