/******************************************************************************
*                                                                             *
* FORMBB - unpack correlations from IFFT.                                     *
*                                                                             *
*     bb_out - pointer to slice containing correlations                       *
*     num    - number of correlation sweeps                                   *
*     p      - 1 >> (p+1) is size of IFFT                                     *
*                                                                             *
******************************************************************************/

#include <cm/paris.h>
#include <math.h>

void FORMBB (bb_out,num,p)
unsigned bb_out, p, num;
{
  unsigned i, pcal, name, save_context_flag;
  unsigned n = 1 << p,size=32;
  CM_geometry_id_t current_geometry;

  pcal =CM_geometry_send_address_length(CM_vp_set_geometry(CM_current_vp_set));
  current_geometry=CM_vp_set_geometry(CM_current_vp_set);
  save_context_flag = CM_allocate_stack_field(1);
  CM_store_context(save_context_flag);
  CM_set_context();
  name = CM_allocate_stack_field(pcal);
  CM_u_move_zero_1L(name, pcal);
  CM_my_send_address_1L(name);

{ unsigned temp;
  unsigned sa,x,y;


  temp=CM_allocate_stack_field(size);
  sa=CM_allocate_stack_field(pcal);
  x=CM_allocate_stack_field(p);
  y=CM_allocate_stack_field(pcal-p);
  CM_my_send_address_1L (sa,pcal);
  CM_extract_news_coordinate_1L(current_geometry,x,0,sa,p);
  CM_u_ge_constant_1L (x,n/4,p);
  CM_logand_context_with_test();
  CM_u_lt_constant_1L (x,n/4+n/2,p);
  CM_logand_context_with_test();

  if (pcal != p) CM_extract_news_coordinate_1L(current_geometry,y,1,sa,pcal-p);
  CM_u_subtract_constant_2_1L(x,n/4,p); 
  CM_u_multiply_constant_2_1L(x,2,p);
  CM_u_move_1L(sa,x,p); 
  if ( pcal != p)  CM_u_move_1L(sa+p,y,pcal-p); 
  for (i=0;i< num;i++) {
    CM_u_move_1L (temp,bb_out+2*size*i,size);
    CM_send_1L ( bb_out+2*size*i,sa,temp,size,CM_no_field);  
    CM_u_move_1L (temp,bb_out+2*size*i+32,size);
    CM_u_add_constant_2_1L(sa,1,p); 
    CM_send_1L ( bb_out+2*size*i,sa,temp,size,CM_no_field);  
    CM_u_subtract_constant_2_1L(sa,1,p); 
   }
  }

  CM_load_context(save_context_flag);
  CM_deallocate_stack_through(save_context_flag);
}
