/*

  gfilt2d.c -- program to convert 501 point Blackman filter impulse response
               stored in front end file to frequency domain data
               stored in parallel DataVault file.  Data is output in bit-
               reversed order for input to bb program.
               Calling sequence is

        cmattach -g 8192 -w gfilt8k infile outfile   &

    where
               infile is the serial file  with the impulse response
               outfile is the parallel DataVault file to contain the 
               transformed data.

*/

#include <sys/types.h>
#include <cm/cm.h>
#include <cm/cm_file.h>
#include <cm/cmfs.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <cm/paris.h>
#include <cm/cmssl-paris.h>

void main(argc, argv)
     int argc;
     char **argv;
{
  char dv_in_file_name[128], dv_out_file_name[128];
  int dv_infile, dv_outfile;
  int debu, dbu,offset[2],dimvec[2];
  int pos,bits,axis[2];
  FILE *files;
  char fchar[128];
  float filt1;
  unsigned sa;
  unsigned  fft_filt;
  unsigned temp;
  unsigned save_context_flag;
  unsigned pcal,p;
  unsigned numb,numb_send;
  CM_geometry_id_t new_geometry;
  unsigned dimension_array[2], new_vp_set;
  CMSSL_fft_setup_t id;
  CMSSL_op_t ops[2]; 
  CMSSL_bit_order_t sbo[2],dbo[2];
  CMSSL_scale_t sca[2];

  struct CM_axis_descriptor axe_desc[32];
  CM_axis_descriptor_t axe_desc_t[32];
  int i,j;

  unsigned start[32],stop[32];
  start[0]=0;start[1]=0;stop[0]=200;stop[1]=3;

  strcpy(dv_in_file_name, argv[1]);
  strcpy(dv_out_file_name, argv[2]);

  CM_init ();
  CMSSL_init ();
  CM_set_safety_mode (1); 
  CM_reset_timer(0);

  ops[0]=CMSSL_f_xform;
  sbo[0]=CMSSL_normal;
  dbo[0]=CMSSL_normal;
  sca[0]=CMSSL_noscale;
  ops[1]=CMSSL_nop;
  sbo[1]=CMSSL_normal;
  dbo[1]=CMSSL_normal;
  sca[1]=CMSSL_noscale;
   for (i = 0; i < 2; i++) {
        axe_desc[i].weight = (1-i)*10;
        axe_desc[i].ordering = CM_send_order;
        axe_desc[i].off_chip_bits = 0;
        axe_desc[i].off_chip_pos = 0; 
        axe_desc[i].on_chip_bits = 0;
        axe_desc[i].on_chip_pos = 0;
        axe_desc_t[i] = &(axe_desc[i]);
      }
   
  p=12;
  pcal=13;
  axe_desc[0].length = 1<< (p-1);
  axe_desc[1].length = 1<< (pcal-p+1);
  new_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]), 2);
  new_vp_set = CM_allocate_vp_set(new_geometry);
  CM_set_vp_set(new_vp_set);
  CM_set_context();
  
  numb = CM_allocate_stack_field (pcal) ;
  numb_send = CM_allocate_stack_field (pcal) ;
  CM_u_move_zero_1L (numb,pcal);
  CM_my_send_address_1L(numb_send);

  fft_filt=CM_allocate_stack_field(64);    
  temp =CM_allocate_stack_field(64);
    
  id= CMSSL_c_c_fft_setup_1L (fft_filt,23,8);
    
  if ((files = fopen(dv_in_file_name, "r")) == NULL) {
    printf("\nError:\tcouldn't open file %s.\n", dv_in_file_name);
    exit(1);
  }
  if ((dv_outfile = CMFS_open(dv_out_file_name, CMFS_O_RDWR | CM_CREAT, 0644,
			      CM_physical_processors_limit,
			      CM_user_number_of_processors_limit)) == -1){
    printf("error on dv file open: %s\n", dv_out_file_name);
    exit(1);
  }

  for (i = 0; i < 250; i++){
    fscanf(files, "%s\n", fchar);
    filt1=atof(fchar);
    for (j=0;j< (1<< (pcal-p+1)); j++) {
      sa=CM_fe_make_news_coordinate ( new_geometry,0,i);
      sa=CM_fe_deposit_news_coordinate (new_geometry,sa,1,j);
      CM_f_write_to_processor (sa,fft_filt,filt1,23,8);
  }
    fscanf(files, "%s\n", fchar);
    filt1=atof(fchar);
    for (j=0;j< (1<< (pcal-p+1)); j++) {
      sa=CM_fe_make_news_coordinate ( new_geometry,0,i);
      sa=CM_fe_deposit_news_coordinate (new_geometry,sa,1,j);
      CM_f_write_to_processor (sa,fft_filt+32,filt1,23,8);
    }
  }
  CMSSL_c_c_fft_detailed_a_1L (fft_filt,id,ops,sbo,dbo,sca,23,8);
  UNS2REAL2D(fft_filt,  p, 1,1,new_geometry);
  TAKE_AMPLITUDE (temp, fft_filt);
  save_context_flag = CM_allocate_stack_field(1);
  CM_store_context(save_context_flag);
  CM_extract_news_coordinate_1L (new_geometry,numb,0,numb_send,pcal) ;
  CM_u_eq_zero_1L (numb+1,p-2) ;
  CM_logand_context_with_test()  ;
  CM_f_move_zero_1L(temp,23,8);
  CM_load_context (save_context_flag) ;
  bits = CMFS_write_file_always(dv_outfile, temp, 32);
  if (bits == -1){CMFS_perror("Had trouble saving set"); exit(1);}

    
      
}

