/*

  LOADDATA - Loads the next set of data from a buffer in CM memory. 
             If some or all of the data is not already in CM memory the data
             is retrieved from the data vault.  Calling sequence is

     LOADDATA ( pure_data,old_data,move,dv_infile,offset,size,sweeps,bitrev)

  where
 
    pure_data - Pointer to slice in CM memory to store data
    old_data  - Pointer to slice in CM memory for first buffer
    move      - Size (number of bits per processor) of each data buffer
    dv_infile - Pointer to data vault vault file
    offset    - Offset into double buffer indicating the first sample 
                to be retrieved
    size     -  Number of samples to be retrieved      
    sweeps   -  Number of sweeps less one in buffer
    bitrev   -  Flag to indicate whether or not data is to bitreversed ordered
*/

#include <cm/fast-graph.h>
#include <cm/paris.h>

  void LOADDATA (pure_data, old_data,move,dv_infile,offset,size,sweeps,bitrev)
  unsigned pure_data, old_data;
  int dv_infile;
  int move,offset,size,sweeps,bitrev;

{
  int pcal,p, n, ntemp,  bits, nsweep,off;
  int i1,i2,channel;
  unsigned name, name_send, save_context_flag,temp,x,y;
  CM_geometry_id_t current_geometry;
  unsigned start[2],stop[2];
  static unsigned last_value_of_p = 0;
  static char *digit[] = { "0","1","2","3","4","5","6","7","8","9" };
  static char out_file[80]=" this is a test";
  static CMFG_trace_id_t trace;

  CM_timer_start(1);
  current_geometry=CM_vp_set_geometry(CM_current_vp_set);
  pcal = CM_geometry_send_address_length(current_geometry);
  p = CM_geometry_coordinate_length(current_geometry,0);
  n = 1 << p;
  save_context_flag = CM_allocate_stack_field(1);
  CM_store_context(save_context_flag);
  CM_set_context();
  name = CM_allocate_stack_field(pcal+1);
  x    = CM_allocate_stack_field(pcal+1);
  y    = CM_allocate_stack_field(pcal);
  name_send = CM_allocate_stack_field(pcal);
  CM_u_move_zero_1L(name,pcal+1);
  CM_u_move_zero_1L(x,pcal+1);
  CM_my_send_address_1L(name_send);
  CM_extract_news_coordinate_1L(CM_vp_set_geometry(CM_current_vp_set),
               name,0,name_send,pcal);
  CM_u_move_zero_1L (pure_data,move);
  CM_extract_news_coordinate_1L(current_geometry,y,1,name_send,pcal);

/* If all of the required data is not in the data vault update buffers  */

  if (offset >= n*sweeps) {
    CM_u_move_1L (old_data,old_data+sweeps*move,move);
    bits = CMFS_read_file(dv_infile, old_data+move, sweeps*move);
    if (bits == -1){
     CMFS_perror("Had trouble reading data."); exit(1);
    }
    offset += -n*sweeps;
   }
  nsweep=offset/n;
  off=offset-nsweep*n;
  
if ( off+ size <= n) {
/* Retrieve data when offset is 0 ( NO inter PE mocves required) */
   if ( off == 0) {
    CM_u_lt_constant_1L(name, size, pcal+1);
    CM_logand_context_with_test();
    CM_u_move_1L (pure_data,old_data+nsweep*move,move);
    CM_timer_stop(1);
    CM_timer_start(2);
    if ( bitrev == 1 ) BITREV2D(pure_data,p,move,0);
    CM_timer_stop(2);
    CM_timer_start(1);
   }
  else {
/* Retrieve data when offset is not zero but data is all within one buffer */
   CM_u_subtract_constant_3_1L (x,name, off, pcal);
   CM_u_lt_constant_1L(x, size, pcal+1);
   CM_logand_context_with_test();
   CM_timer_stop(1);
   CM_timer_start(2);
   if (bitrev == 1 ) CM_move_reversed_1L (x, x, p);
   CM_make_news_coordinate_1L ( current_geometry,name,0,x,p);
   CM_deposit_news_coordinate_1L (current_geometry,name,1,y,pcal-p);
   CM_send_1L(pure_data,name,old_data+nsweep*move,move);
   CM_timer_stop(2);
   CM_timer_start(1);
  }
 }

  else {
/* Retrieve data when the data is in both buffers ( different slices) */
  if (pcal!= last_value_of_p) {
    last_value_of_p=pcal;
  strcpy (out_file,"FG/loadtrace");
  channel=CM_geometry_coordinate_length(current_geometry,0);
  i1=channel/10;
  i2=channel-i1*10;
  strcat (out_file,digit[i1]);
  strcat (out_file,digit[i2]);
  channel=CM_geometry_coordinate_length(current_geometry,1);
  strcat (out_file,digit[channel]);
  if ( bitrev == 1) strcat (out_file,"b");
  printf (" string  = %s \n", out_file);
  trace=CMFG_dvpath_read_trace(out_file);
  }
   temp=CM_allocate_stack_field(move);
   CM_u_ge_constant_1L(name, off, pcal);
   CM_logand_context_with_test();
   CM_u_move_1L ( temp, old_data+nsweep*move, move);
   CM_set_context();
   CM_u_lt_constant_1L(name, off-n+size, pcal);
   CM_logand_context_with_test();
   CM_u_move_1L ( temp, old_data+(nsweep+1)*move, move);
   CM_set_context();
   CM_u_subtract_constant_3_1L (x, name, off, p);
   CM_u_lt_constant_1L(x, size, pcal+1);
   CM_logand_context_with_test();
   CM_timer_stop(1);
   CM_timer_start(2);
   CMFG_deliver_1L(pure_data,trace,temp,move);
  CM_timer_stop(2);
  CM_timer_start(1);
  }

  CM_load_context(save_context_flag);
  CM_deallocate_stack_through(save_context_flag);
  CM_timer_stop(1);
 }
