/******************************************************************************
*                                                                             *
*  NEGODDC - Negates the complex values in a slice of CM memory for PEs       *
*            that correspond to odd FFT ccoefficients.  This is the odd       *
*            PEs if the data in the slice is in normal order and the upper    *
*            half of the PEs (when grouped by FFT) if the data is arranged    *
*            in bit reversed order.                                           *
*                                                                             *
*       data   - Slice of complex data in CM memory                           *
*       number - Number of complex values in slice                            *
*       bitrev - Flag that indicates whther or not data is in normal order    *
*                                                                             *
******************************************************************************/

#include <cm/paris.h>

void NEGODDC (data,number,bitrev)
unsigned data;
int number,bitrev;

{
  int pcal,i,p;
  unsigned name,name_send, save_context_flag;
  CM_geometry_id_t default_geometry_id;

  default_geometry_id = CM_vp_set_geometry(CM_current_vp_set);
  pcal = CM_geometry_send_address_length(default_geometry_id);
  p = CM_geometry_coordinate_length(default_geometry_id,0);
  save_context_flag = CM_allocate_stack_field(1);
  CM_store_context(save_context_flag);
  CM_set_context();
  name_send = CM_allocate_stack_field(pcal);
  name = CM_allocate_stack_field(pcal);
  CM_my_send_address_1L(name_send);
  CM_u_move_zero_1L (name,pcal);
  CM_extract_news_coordinate_1L(default_geometry_id,name,0,name_send,pcal);
 
  if ( bitrev == 0) {
    CM_logand_context (name);
   }
  else {
    CM_logand_context (name+p-1);
   }

  for (i=0;i< number; i++){
    CM_f_negate_1_1L (data+i*64,23,8);
    CM_f_negate_1_1L (data+32+i*64,23,8);
   }

  CM_load_context(save_context_flag);
  CM_deallocate_stack_through(save_context_flag);
}
 
