/******************************************************************************
*                                                                             *
*    SCR2REAL2D: Performs the scramble step for doing an inverse  N point     *
*              complex-to-real FFTusing a N/2 point complex FFT (inverse      *
*              operation of Brigham, The Fast Fourier Transform pp 167-169).  *
*                                                                             *
*        real_part -  Pointer to data on CM                                   *
*        n         -  Log base 2 of the number of real data points            *
*        num       -  Number of FFTs to be scramble                           *
*        unpack    -  Flag indicating whether or not Nyquist value has been   *
*                     into imaginery part of DC                               *
*        bitrev    -  Flag indicating whther or not data is in bit reversed   *
*                     order                                                   *
*                                                                             *
******************************************************************************/

#include <cm/paris.h>
#include <math.h>

  void SCR2REAL2D (real_part, p, num,unpack,bitrev)
  unsigned real_part,  p, num,unpack,bitrev;
{

  unsigned i,j,j32,n, n2 ;
  unsigned  real_assoc, temp;
  unsigned real_temp;
  static unsigned numb,numb_send,real_cos,real_sin ;
  static  unsigned sa,x2,y,pprev=0;
  static CM_geometry_id_t default_geometry_id;
  static unsigned last_value_of_p = 0;
  unsigned pcal, name, save_context_flag,flag;

  p= p-1;
  n= 1 << p;
  default_geometry_id = CM_vp_set_geometry(CM_current_vp_set);
  pcal = CM_geometry_send_address_length(default_geometry_id);
  if (pcal!= last_value_of_p) {
    last_value_of_p = pcal;
    sa=CM_allocate_stack_field(pcal);
    y=CM_allocate_stack_field(pcal);
    x2=CM_allocate_stack_field(pcal);
    real_sin = CM_allocate_stack_field(32);
    real_cos = CM_allocate_stack_field(32);
    numb = CM_allocate_stack_field (pcal) ;
    numb_send = CM_allocate_stack_field (pcal) ;
    CM_u_move_zero_1L (numb,pcal);
    CM_my_send_address_1L(numb_send);
    CM_extract_news_coordinate_1L (default_geometry_id,numb,0,numb_send,pcal) ; 
    if ( bitrev == 1) CM_move_reversed_1L (numb, numb, p);
    CM_f_u_float_2_2L (real_sin,numb,p,23,8);
    CM_f_multiply_constant_2_1L (real_sin,3.1415926/n,23,8);
    CM_f_cos_2_1L (real_cos,real_sin,23,8);
    CM_f_sin_1_1L (real_sin,23,8);
   }
   real_assoc = CM_allocate_stack_field(64*num);
   real_temp = CM_allocate_stack_field(64*num);
   temp = CM_allocate_stack_field(64);
   CM_u_move_zero_1L (numb,pcal);
   CM_my_send_address_1L (numb) ;
   if (pprev!=p) {
    pprev = p;
    CM_u_move_zero_1L (x2,pcal);
    CM_u_move_zero_1L (y,pcal);
    CM_my_send_address_1L (sa);
    CM_extract_news_coordinate_1L(default_geometry_id,x2,0,sa,pcal);
    if ( bitrev == 1) CM_move_reversed_1L (x2, x2, p);
    CM_u_subtract_3_1L (x2,y,x2,p);
    if (bitrev == 1 ) CM_move_reversed_1L (x2, x2, p);
    CM_u_move_1L (y,sa+p,pcal-p);
    CM_extract_news_coordinate_1L(default_geometry_id,y,1,sa,pcal);
    CM_make_news_coordinate_1L ( default_geometry_id,sa,0,x2,pcal);
    CM_deposit_news_coordinate_1L (default_geometry_id,sa,1,y,pcal);
    CM_my_send_address_1L (y);
    CM_send_1L (x2,sa,y,pcal,CM_no_field);
  }      
  CM_u_move_1L (real_assoc, real_part, 64*num) ;
  CM_send_1L(real_temp, x2, real_part, 64*num, CM_no_field);
  for ( j=0; j< 64*num ; j+= 64) {
    j32=j+32;
    CM_f_add_mult_const_1L(temp+32,real_assoc+j32,real_temp+j32,.5,23,8);
    CM_f_sub_mult_const_1L(temp,real_assoc+j,real_temp+j,.5,23,8);
    CM_f_add_mult_const_1L(real_part+j,real_assoc+j,real_temp+j,.5,23,8);
    CM_f_mult_subf_1L (real_part+j,real_cos,temp+32,real_part+j,23,8);
    CM_f_mult_subf_1L (real_part+j,real_sin,temp,real_part+j,23,8);
    CM_f_sub_mult_const_1L(real_part+j32,real_assoc+j32,real_temp+j32,.5,23,8);
    CM_f_mult_subf_1L (real_part+j32,real_sin,temp+32,real_part+j32,23,8);
    CM_f_mult_add_1L (real_part+j32,real_cos,temp,real_part+j32,23,8);
  }
  if (unpack != 0) {
    save_context_flag = CM_allocate_stack_field(1);
    CM_store_context(save_context_flag);
    CM_extract_news_coordinate_1L (default_geometry_id,
                  numb,0,numb_send,pcal) ; 
    CM_u_eq_zero_1L (numb,p) ;
    CM_logand_context_with_test()  ;
    for ( j=0; j< 64*num ; j+= 64) {
      j32=j+32;
      CM_f_add_mult_const_1L(real_part+j,real_temp+j,real_temp+j32,.5,23,8);
      CM_f_sub_mult_const_1L(real_part+j32,real_temp+j,real_temp+j32,.5,23,8);
    }
    CM_load_context (save_context_flag) ;
   }
  CM_deallocate_stack_through(real_assoc); 
}
  




