/******************************************************************************
*                                                                             *
*  sertopar.c - Remultiplex a serial DataVault file aad write it out to       *
*               a parallel DataVault file. Running sequence is                *
*                                                                             *
*    sertopar in out -b n1 -m n2 -l n3 n4 -h n5 n6 -z n7                      *
*                                                                             *
*       where all - switches are optional and order independent               *
*                                                                             *
*         in       -  Input DataVault file                                    *
*         out      -  Output DataVault file                                   *
*         -b n1    -  Number of bytes comprising a time-series sample         *
*         -m n2    -  Number of channels of data                              *
*         -l n3 n4 -  Sub-array 1 channel numbers                             *
*         -h n5 n6 -  Sub-array 2 channel numbers                             *
*         -z n7    -  Number of bytes in serial input file                    *
*                                                                             *
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <cm/paris.h>
#include <cm/cm_file.h>
#include <cm/cmfs.h>
#include <cm/cm_stat.h>

void main(argc, argv)
int argc;
char **argv;
{

  char dv_in_file[128],dv_out_file[128];
  unsigned in_data, out_data;
  int bits_in,bits_out,ptr;
  int i,j,k,kk,move,bytes=4;
  int pts;
  int channels=8;
  int number_bytes =0;
  int steps,buffer,inner,offset;
  int dvfile_in,dvfile_out;
  int lo1,lo2,hi1,hi2 ;
  long transpose_size=16384/2/4;

  if (argc < 3)
    {printf("useage: reformat in_filename out_filename.\n"); exit(0);}
  strcpy(dv_in_file, argv[1]);
  strcpy(dv_out_file, argv[2]);
  for (ptr=3; ptr<argc; ptr++) {
    char *arg = argv[ptr];
    if (arg[0] == '-') {
      switch (arg[1]) {
      case 'm':                          
        if (++ptr >= argc) exit(1);
        channels = atoi(argv[ptr]);
        continue;
      case 'b':                          
        if (++ptr >= argc) exit(1);
        bytes = atoi(argv[ptr]);
        continue;
      case 'l':                          
        if (++ptr >= argc) exit(1);
        lo1 = atoi(argv[ptr]);
        if (++ptr >= argc) exit(1);
        lo2 = atoi(argv[ptr]);
        continue;
      case 'h':                          
        if (++ptr >= argc) exit(1);
        hi1 = atoi(argv[ptr]);
        if (++ptr >= argc) exit(1);
        hi2 = atoi(argv[ptr]);
        continue;
      case 'z':                          
        if (++ptr >= argc) exit(1);
        number_bytes = atoi(argv[ptr]);
        continue;
      default: exit(1);
      }
    }
    else exit(1);
  }

  CM_init();
  CM_set_safety_mode (0);
  CM_set_context();
  pts=  
   1<<CM_geometry_send_address_length(CM_vp_set_geometry(CM_current_vp_set));

  in_data=CM_allocate_stack_field(transpose_size);
  out_data=CM_allocate_stack_field(transpose_size);

  printf("opening in file...\n");
  if ((dvfile_in = CMFS_open(dv_in_file, CMFS_O_RDONLY, 0644))  == -1){
    printf("error on dv file open: %s\n", dv_in_file);
    exit(1);  }
  printf("opening out file...\n");
  if ((dvfile_out = CMFS_open(dv_out_file, 
     CMFS_O_WRONLY | CMFS_O_CREAT | CMFS_O_APPEND, 0644))  == -1){
    printf("error on out file open: %s\n", dv_out_file);
    exit(1);  }
  bits_in=transpose_size;
  while (bits_in == transpose_size) {
    bits_in = CMFS_read_file(dvfile_in, in_data, transpose_size);
    if (bits_in == -1){CMFS_perror("Couldn't read some data"); exit(-1);}
    else { printf("read %d bits of data  per vp.\n",
	      bits_in);
       }
    offset =0;
    for (j=0;j<bits_in; j+= 8* bytes * channels){
      for (i=lo1;i<lo2+1;i++) {
        CM_u_move_1L (out_data+offset,in_data+j+i*8*bytes,8*bytes);
        offset += 8*bytes;
       }
      for (i=hi1;i<hi2+1;i++) {
        CM_u_move_1L (out_data+offset,in_data+j+i*8*bytes,8*bytes);
        offset += 8*bytes;
       }
     }
    bits_out = CMFS_write_file(dvfile_out, out_data,offset);
    if (number_bytes != 0) {
      number_bytes -= (transpose_size/8)*pts;
      if (number_bytes <= 0) exit(0);
     }
   }
}
