
/*
    UNS2REAL: Performs the unscramble step for doing a N point real FFT 
              using a N/2 point complex FFT (see Brigham, The Fast Fourier 
             Transform pages 167-169).
  
   Calling procedure:
       real_part = CM_allocate_stack_field(64*num);
        UNS2REAL2D(real_part,  p, num);

 where

        real_part       is pointer to data on CM
        n               is the log base 2 of the number of real data points
        num             is the number of FFTs to be unscrambled
        geometry        is the geometry of the vp set containing the data
                        to be FFT'd.
        bitrev          flag indicating whther or not data is in bit reversed
                        order
*/

#include <cm/fast-graph.h>
#include <cm/paris.h>
#include <math.h>

  void UNS2REAL2D (real_part,  p, num,pack,bitrev)
  unsigned real_part,  p, num,pack,bitrev;
{

  unsigned i,j,j32,n, n2,i1,i2,channel ;
  unsigned  real_assoc, temp;
  unsigned real_temp;
  static unsigned numb,numb_send,real_cos,real_sin ;
  static  unsigned sa,x2,y,pprev=0;
  static CM_geometry_id_t default_geometry;
  static unsigned last_value_of_p = 0;
  unsigned pcal, name, save_context_flag,flag;
  static char *digit[] = { "0","1","2","3","4","5","6","7","8","9" };
  static char out_file[80]=" this is a test";
  static CMFG_trace_id_t trace;

/* ---------------- BEGIN GEOMETRY CONFIGURATION ---------------- */
  p= p-1;
  n= 1 << p;
  default_geometry = CM_vp_set_geometry(CM_current_vp_set);
  pcal = CM_geometry_send_address_length(default_geometry);
  if (pcal!= last_value_of_p) {
   last_value_of_p = pcal;
   sa=CM_allocate_stack_field(pcal);
   y=CM_allocate_stack_field(pcal);
   x2=CM_allocate_stack_field(pcal);
   real_sin = CM_allocate_stack_field(32);
   real_cos = CM_allocate_stack_field(32);
   numb = CM_allocate_stack_field (pcal) ;
   numb_send = CM_allocate_stack_field (pcal) ;
   CM_u_move_zero_1L (numb,pcal);
   CM_my_send_address_1L(numb_send);
   CM_extract_news_coordinate_1L (default_geometry,numb,0,numb_send,pcal) ; 
   if ( bitrev == 1 ) CM_move_reversed_1L (numb, numb, p);
   CM_f_u_float_2_2L (real_sin,numb,p,23,8);
   CM_f_multiply_constant_2_1L (real_sin,3.1415926/n,23,8);
   CM_f_cos_2_1L (real_cos,real_sin,23,8);
   CM_f_sin_1_1L (real_sin,23,8);
 }
 real_assoc = CM_allocate_stack_field(64*num);
 real_temp = CM_allocate_stack_field(64*num);
 temp = CM_allocate_stack_field(64);
 CM_u_move_zero_1L (numb,pcal);
 CM_my_send_address_1L (numb) ;
 if (pprev!=p) {
  pprev = p;
  strcpy (out_file,"FG/unstrace");
  channel=CM_geometry_coordinate_length(default_geometry,0);
  i1=channel/10;
  i2=channel-i1*10;
  strcat (out_file,digit[i1]);
  strcat (out_file,digit[i2]);
  channel=CM_geometry_coordinate_length(default_geometry,1);
  strcat (out_file,digit[channel]);
  printf (" string  = %s \n", out_file);
  trace=CMFG_dvpath_read_trace(out_file);
 }    
  CM_u_move_1L (real_assoc, real_part, 64*num) ;
  CMFG_deliver_1L(real_temp, trace, real_part, 64*num);
  for ( j=0; j< 64*num ; j+= 64) {
    j32=j+32;
    CM_f_add_mult_const_1L(temp+32,real_assoc+j32,real_temp+j32,.5,23,8);
    CM_f_sub_mult_const_1L(temp,real_assoc+j,real_temp+j,.5,23,8);
    CM_f_add_mult_const_1L(real_part+j,real_assoc+j,real_temp+j,.5,23,8);
    CM_f_mult_add_1L (real_part+j,real_cos,temp+32,real_part+j,23,8);
    CM_f_mult_subf_1L (real_part+j,real_sin,temp,real_part+j,23,8);
    CM_f_sub_mult_const_1L(real_part+j32,real_assoc+j32,real_temp+j32,.5,23,8);
    CM_f_mult_subf_1L (real_part+j32,real_sin,temp+32,real_part+j32,23,8);
    CM_f_mult_subf_1L (real_part+j32,real_cos,temp,real_part+j32,23,8);
  }

  if (pack != 0) {
    save_context_flag = CM_allocate_stack_field(1);
    CM_store_context(save_context_flag);
    CM_extract_news_coordinate_1L (default_geometry,numb,0,numb_send,pcal) ; 
    CM_u_eq_zero_1L (numb,p) ;
    CM_logand_context_with_test()  ;
    for ( j=0; j< 64*num ; j+= 64) {
      j32=j+32;
      CM_f_subtract_3_1L(real_part+j32,real_temp+j,real_temp+j32,23,8);
    }
   CM_load_context (save_context_flag) ;
  }
  CM_deallocate_stack_through(real_assoc); 
}
  




