#define current Current
#include <geodata_private.h>
#undef current
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

/* This module demonstrates the ease of using geodata routines with CM
   data sets. The data sets used are generated by the user as C* parallel
   variables. */


float:current smooth(float:current in)
{
  float:current temp = in;
  [.-1][.]temp+=in;
  [.+1][.]temp+=in;
  [.][.-1]temp+=in;
  [.][.+1]temp+=in;
  return (temp/5.0);
}


shape [128][128]weather;

main(argc,argv)
     int argc;
     char **argv;
{
  int hx,hy,hdx,hdy;
  int lx,ly,ldx,ldy;
  int i,j,steps;
 int:weather low_v,low_u;
 float:weather pressure,wind;
 double:weather wind_low;
  GB_id pressure_data,wind_data,scene,display;
  CM_field_id_t context;

  with(weather) {
    
    /* select one in 64 points to show wind direction arrows */
    low_u=pcoord(0);
    low_v=pcoord(1);
    context=CM_allocate_heap_field(1);
    where(low_u%8==0 && low_v%8==0) {
      CM_store_context(context);
    }

    /* start up Geodata and read the gl file describing the data sets,
       representations, scene, and display */

    if (argc == 1)
      GI_init("sample_2.gl");
    else
      GI_init(argv[1]);
    scene=GI_get_scene_id("sample_2_scene");
    display=GI_get_display_id("sample_2_display");

    /* tie the CM data sets to the data set descriptors used by Geodata */
    pressure_data=GI_get_data_id("sample_2_pressure");
    wind_data=GI_get_data_id("sample_2_wind");
    GI_set_cm_data(pressure_data,&pressure,0);
    GI_set_cm_data(wind_data,&wind,context);

    pressure=0.0;

    hx=10;hy=50;
    hdx=3;hdy=1;
    lx=98;ly=118;
    ldx=-2;ldy=-3;

    /* pump some initial enegy into the weather system */
    for(i=0;i<8;i++) {
      [hx][hy]pressure +=  10000.0;
      [lx][ly]pressure += -10000.0;
      for(j=0;j<16;j++)pressure=smooth(pressure);
    }
    printf("%f %f\n",<?=pressure,>?=pressure);

    steps=30;
    for(i=0;i<steps;i++) {
      everywhere {

	/* put high and low pressure points into the system and spread
	   the energy around to simulate high and low pressure zones */
	[hx][hy]pressure +=  10000.0;
	[lx][ly]pressure += -10000.0;
	for(j=0;j<16;j++)pressure=smooth(pressure);

	/* have each proccessor calculate a vector towarc the low
	   pressure point and a vector away form the high pressure point.
	   sum the vectors. */
	low_u=pcoord(0);
	low_v=pcoord(1);
	low_u = low_u - 128/lx + 128/hx;
	low_v = low_v - 128/ly + 128/hy;

	/* convert the vector to angular measurements */
	wind_low = (double:current) (low_u * low_u + low_v * low_v);
	wind_low = sqrt(wind_low);
	wind_low = ((double:current)low_v) / wind_low;
	wind_low = acos(wind_low);
	where(low_u < 0) wind_low = -wind_low;
	where(wind_low > GS_PI) wind_low = wind_low - GS_PI;
	where(wind_low < -GS_PI) wind_low = wind_low + GS_PI;
	wind=(float:current)wind_low;
      }
      /* render and display the scene */
      GI_update_scene(scene);
      GI_display_scene(display,scene,0,0);

      /* move the high and low pressure points */
      hx += hdx; hy += hdy;
      lx += ldx; ly += ldy;
    }
  }
}
