#include <cm/paris.h>
#include <math.h>
#include <geodata.h>
#include <geodata_private.h>
#include <pgl/pgl.h>

#define FROM_0 0
#define FROM_1 1
#define FROM_2 2
#define TO_0   3
#define TO_1   4
#define TO_2   5
#define UP_0   6
#define UP_1   7
#define UP_2   8
#define FOV    9
#define HITHER 10
#define YON   11


#define EDGE 1.4

/* from[0] from[1] from[2] to[0] to[1] to[2] fov twist hither yon */
void GD_generate_xform (dest,parameters)
     float dest[4][4];
     float parameters[NO_PROJECTION_PARAMS];
{
  float temp[4][4];
  float vup[3],vpn[3],vrp[3],t1[3],t2[3],t3[3],zmin;
  float umin,umax,vmin,vmax,d;
  int i;
/*
  vup[1]=vup[0]=0.0;
  vup[2]=(-1.0);
  PGL_identity (temp);
  PGL_rotate (temp,PGL_z,(M_PI*parameters[PITCH]/180.0));
  PGL_rotate (temp,PGL_x,(M_PI*parameters[YAW]/180.0));
  PGL_fe_transform (vup,vup,temp);
*/

  umin = sqrt(parameters[UP_0]*parameters[UP_0]+
	      parameters[UP_1]*parameters[UP_1]+
	      parameters[UP_2]*parameters[UP_2]);

  vup[0]=(-parameters[UP_0]/umin);
  vup[1]=(-parameters[UP_1]/umin);
  vup[2]=(-parameters[UP_2]/umin);

  PGL_identity (dest);
  PGL_identity (temp);
  temp[3][0] = (-parameters[FROM_0]);
  temp[3][1] = (-parameters[FROM_1]);
  temp[3][2] = (-parameters[FROM_2]);
  PGL_compose (dest,temp);

  PGL_identity (temp);
  t3[0] = (-(vpn[0] = parameters[FROM_0]-parameters[TO_0]));
  t3[1] = (-(vpn[1] = parameters[FROM_1]-parameters[TO_1]));
  t3[2] = (-(vpn[2] = parameters[FROM_2]-parameters[TO_2]));
  
  PGL_normalize(t3);  
  temp[0][2]=t3[0];
  temp[1][2]=t3[1];
  temp[2][2]=t3[2];
  
  PGL_cross(t1,t3,vup);
  PGL_normalize(t1);
  temp[0][0]=t1[0];
  temp[1][0]=t1[1];
  temp[2][0]=t1[2];

  PGL_cross(t2,t1,t3);
  PGL_normalize(t2);
  temp[0][1]=t2[0];
  temp[1][1]=t2[1];
  temp[2][1]=t2[2];
  
  PGL_compose (dest,temp);

  PGL_fe_transform (vrp,parameters+TO_0,dest);
  
  umin=vmin=(-(umax=vmax=vrp[2]*sin(M_PI*parameters[FOV]/180.0)/
	       sin(M_PI*(0.5-parameters[FOV]/180.0))));

  PGL_identity (temp);
  temp[2][0]=(-(vrp[0]+0.5*(umin+umax)))/vrp[2];
  temp[2][1]=(-(vrp[1]+0.5*(vmin+vmax)))/vrp[2];
  PGL_compose (dest,temp);

  PGL_identity (temp);
  temp[0][0]=2.0*vrp[2]/((umax-umin)*parameters[YON]);
  temp[1][1]=2.0*vrp[2]/((vmax-vmin)*parameters[YON]);
  temp[2][2]=1.0/parameters[YON];
  PGL_compose (dest,temp);

  zmin = parameters[HITHER]/parameters[YON];
  PGL_identity (temp);
  temp[2][2]=1.0/(1.0-zmin);
  temp[2][3]=1.0;
  temp[3][2]=(-zmin)/(1.0-zmin);
  temp[3][3]=0.0;
  PGL_compose (dest,temp);
  
  PGL_translate (dest,PGL_z,-0.5);
  PGL_scale (dest,PGL_y,-1.0);
  PGL_scale (dest,PGL_z,-2.0);
}


void GD_fe_general_perspective(data, success, parameters)
     float data[3];
     int *success;
     float *parameters;
{
  float xform[4][4];
  GD_generate_xform(xform,parameters);
  PGL_fe_transform (data,data,xform);
  if ((data[0] > -EDGE) && (data[0] < EDGE) &&
      (data[1] > -EDGE) && (data[1] < EDGE) &&
      (data[2] > -EDGE) && (data[2] < EDGE))
    *success = 1;
  else 
    success = 0;
}


void GD_fe_inverse_general_perspective(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  printf ("No general_perspective inverse!\n");
}


void GD_general_perspective (data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  float xform[4][4];
  GD_generate_xform(xform,parameters);

  PGL_transform (data,data,xform,32);

  CM_u_move_zero_always_1L (success,0,1);

  CM_f_gt_constant_1L (data,-EDGE,23,8);
  CM_store_test (success);
  CM_f_lt_constant_1L (data,EDGE,23,8);
  CM_logand_2_1L (success,CM_test_flag,1);

  CM_f_gt_constant_1L (data+32,-EDGE,23,8);
  CM_logand_2_1L (success,CM_test_flag,1);
  CM_f_lt_constant_1L (data+32,EDGE,23,8);
  CM_logand_2_1L (success,CM_test_flag,1);


  CM_f_gt_constant_1L (data+64,-EDGE,23,8);
  CM_logand_2_1L (success,CM_test_flag,1);
  CM_f_lt_constant_1L (data+64,EDGE,23,8);
  CM_logand_2_1L (success,CM_test_flag,1);

}


void GD_inverse_general_perspective(data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  printf ("No general_perspective inverse!\n");
}
