#include <geodata_private.h>

void GD_map();

struct component {
  char *name;
  char *(*offset_function)();
  int (*compare_function)();
  int array_length;
  void (*map_function)();
};


int compare_int(a,b)
     int *a,*b;
{
 return ((*a ) == (*b));
}


int compare_float(a,b)
     float *a,*b;
{
 return ((*a ) == (*b));
}




/* for the moment this assumes that no resampling will be needed*/

void GD_regrid (dest,source)
     GD_data_desc *dest,*source;
{
  struct GD_data_desc source_temp;
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  CM_field_id_t data,address,length;
  float parameters[10];
  unsigned block_dim[32];

  GD_open_ds(source,GS_NO_PROJECTION,parameters,block_dim);
  while ((data=GD_read_ds(source_temp,GS_NO_PROJECTION,parameters,block_dim))
	 !=CM_no_field){

    
    CM_send_1L (dest->values,source_temp.values,address,length,CM_no_field);
  }
  GD_close_ds (source);
  CM_set_vp_set (entry_vp_set);
}
  
/* extern void map_fe_to_cm(); */

extern char *GB_data_desc_yp_accessor();
extern char *GB_data_desc_user_yp_accessor();
extern char *GB_data_desc_nv_accessor();
extern char *GB_data_desc_r_accessor();
extern char *GB_data_desc_rank_accessor();
extern char *GB_data_desc_proj_fn_accessor();
extern char *GB_data_desc_type_accessor();
extern char *GB_data_desc_units_accessor();
extern char *GB_data_desc_coord_syst_accessor();
extern char *GB_data_desc_filename_accessor();
extern char *GB_data_desc_detail_accessor();
extern char *GB_data_desc_span_accessor();
extern char *GB_data_desc_file_accessor();
extern char *GB_data_desc_file_origin_accessor();
extern char *GB_data_desc_periodic_axes_accessor();
extern char *GB_data_desc_proj_params_accessor();
extern char *GB_data_desc_extent_accessor();
extern char *GB_data_desc_surface_accessor();
extern char *GB_data_desc_origin_accessor();
extern char *GB_data_desc_load_type_accessor();
extern char *GB_data_desc_geometry_accessor();
/*extern char *GB_data_desc_cm_fe_flag_accessor();*/


static struct component components[]=
{ 
  {"detail",GB_data_desc_detail_accessor,0,0,0},
  {"filename",GB_data_desc_filename_accessor,0,0,0},
  {"coord",GB_data_desc_coord_syst_accessor,compare_int,1,0},
  {"units",GB_data_desc_units_accessor,compare_int,3,0},
  {"type",GB_data_desc_type_accessor,compare_int,1,0},
  {"proj_fn",GB_data_desc_proj_fn_accessor,compare_int,1,0},
  {"rank",GB_data_desc_rank_accessor,compare_int,1,0},
  {"r",GB_data_desc_r_accessor,compare_float,3,0},
  {"nv",GB_data_desc_nv_accessor,compare_float,3,0},
  {"user_yp",GB_data_desc_user_yp_accessor,compare_int,1,0},
  {"yp",GB_data_desc_yp_accessor,compare_float,3,0},
  {"origin",GB_data_desc_origin_accessor,compare_float,3,0},
  {"surface",GB_data_desc_surface_accessor,compare_float,1,0},
  {"extent",GB_data_desc_extent_accessor,compare_float,3,0},
  {"proj_params",GB_data_desc_proj_params_accessor,compare_float,
     NO_PROJECTION_PARAMS ,0},
  {"periodic_axes",GB_data_desc_periodic_axes_accessor,0,0,0},
  {"span",GB_data_desc_span_accessor,compare_int,3,0},
  {"geometry",GB_data_desc_geometry_accessor,compare_int,4,0},
  {"load_type",GB_data_desc_load_type_accessor,compare_int,1,0},
/*{"cm_fe_flag",GB_data_desc_cm_fe_flag_accessor,compare_int,1,map_fe_to_cm},*/
  {0,0,0,0,0}};



/* this routine can allocate heap data*/

void GD_map (to,from)
     GD_data_desc *to,*from;
  
{
  int i,j;
  char message[128];
  char *to_pointer,*from_pointer;
  GD_data_desc new_to;
  int result=0;

  bcopy (to,&new_to,sizeof(GD_data_desc)); /*lovely*/
  for (i=0;components[i].name != 0;i++)
    if (components[i].compare_function){
      result=0;
      to_pointer = (*(components[i].offset_function))(&new_to);
      from_pointer = (*(components[i].offset_function))(from);
      for (j=0;j<components[i].array_length;j++){
	result|=!(*(components[i].compare_function))(to_pointer,from_pointer);
	to_pointer+=4;
	from_pointer+=4;
      }
      if (result)
	if (components[i].map_function)
	  (*(components[i].map_function))(&new_to,from);
	else {
	  sprintf (message,"mapping %s between `%s' and\n `%s' failed;",
		   components[i].name,
		   from->detail,new_to.detail);
	  sprintf (message,"%s no such mapping.",message);
	  GB_message (GB_NOTIFY,"Map data set",message);	  
	}  
    }
}


