

#include "pgl_internal.h"
#include <math.h>

static unsigned axes[] = {PGL_x,PGL_y,PGL_z};

/* think about how to move this to pgl_transform*/
void PGL_scale_buffer(buffer,axis,value) 
     PGL_buffer_t buffer;
     unsigned axis;
     float value;
{
  PGL_scale(buffer->transform,axis,value);
  buffer->transform_t = 1;
}


void PGL_translate_buffer(buffer,axis,value) 
     PGL_buffer_t buffer;
     unsigned axis;
     float value;
{
  PGL_translate(buffer->transform,axis,value);
  buffer->transform_t = 1;
}


void PGL_rotate_buffer(buffer,axis,value) 
     PGL_buffer_t buffer;
     unsigned axis;
     float value;
{
  PGL_rotate(buffer->transform,axis,value);
  buffer->transform_t = 1;
}

void PGL_clear_buffer_transform(buffer)
     PGL_buffer_t buffer;
{
  PGL_identity (buffer->transform);
  buffer->transform_t=0;
}



void PGL_clear_buffer(buffer,r,g,b)
     PGL_buffer_t buffer;
     unsigned r,g,b;
{
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  CM_set_vp_set (buffer->vp_set);
  CM_u_move_zero_always_1L (buffer->field,PGL_point_length(buffer->format));
  CM_set_context();
  PGL_broadcast (PGL_extract(buffer->field,buffer->format,PGL_r),
		 &r,PGL_type(buffer->format,PGL_r),PGL_unsigned,
		 PGL_length(buffer->format,PGL_r),32);
  PGL_broadcast (PGL_extract(buffer->field,buffer->format,PGL_g),
		 &g,PGL_type(buffer->format,PGL_g),PGL_unsigned,
		 PGL_length(buffer->format,PGL_g),32);
  PGL_broadcast (PGL_extract(buffer->field,buffer->format,PGL_b),
		 &b,PGL_type(buffer->format,PGL_b),PGL_unsigned,
		 PGL_length(buffer->format,PGL_b),32);
  CM_set_vp_set(entry_vp_set);

  buffer->old_data_items=buffer->data_items;
  buffer->old_data_item_count=buffer->data_item_count;
  buffer->data_items=0;
  buffer->data_item_count=0;
}


void PGL_buffer_ambient (buffer,r,g,b)
     PGL_buffer_t buffer;
     unsigned r,g,b;
{
  buffer->ambient[0]=r;
  buffer->ambient[1]=g;
  buffer->ambient[2]=b;
}


/* clean this up...aspect ratio needs to be built into u and v*/
void PGL_perspective_view (buffer,from,to,up,fov,back,front)
     PGL_buffer_t buffer;
     float from[3],to[3],up[3];
     float fov, back,front;
{
  float temp[4][4];
  float vpn[3],vrp[3],t1[3],t2[3],t3[3],zmin;
  float umin,umax,vmin,vmax,d;

  PGL_identity (temp);
  temp[3][0] = (-from[0]);
  temp[3][1] = (-from[1]);
  temp[3][2] = (-from[2]);
  PGL_compose (buffer->transform,temp);

  PGL_identity (temp);
  t3[0] = (-(vpn[0] = from[0]-to[0]));
  t3[1] = (-(vpn[1] = from[1]-to[1]));
  t3[2] = (-(vpn[2] = from[2]-to[2]));
  
  PGL_normalize(t3);  
  temp[0][2]=t3[0];
  temp[1][2]=t3[1];
  temp[2][2]=t3[2];
  
  PGL_cross(t1,t3,up);
  PGL_normalize(t1);
  temp[0][0]=t1[0];
  temp[1][0]=t1[1];
  temp[2][0]=t1[2];

  PGL_cross(t2,t1,t3);
  PGL_normalize(t2);
  temp[0][1]=t2[0];
  temp[1][1]=t2[1];
  temp[2][1]=t2[2];
  
  PGL_compose (buffer->transform,temp);

  PGL_fe_transform (vrp,to,buffer->transform);

  umin=vmin=
   (-(umax=vmax=vrp[2]*sin(M_PI*fov/180.0)/sin(M_PI*(0.5-fov/180.0))));

  PGL_identity (temp);
  temp[2][0]=(-(vrp[0]+0.5*(umin+umax)))/vrp[2];
  temp[2][1]=(-(vrp[1]+0.5*(vmin+vmax)))/vrp[2];
  PGL_compose (buffer->transform,temp);

  PGL_identity (temp);
  temp[0][0]=2.0*vrp[2]/((umax-umin)*back);
  temp[1][1]=2.0*vrp[2]/((vmax-vmin)*back);
  temp[2][2]=1.0/back;
  PGL_compose (buffer->transform,temp);

  zmin =front/back;
  PGL_identity (temp);
  temp[2][2]=1.0/(1.0-zmin);
  temp[2][3]=1.0;
  temp[3][2]=(-zmin)/(1.0-zmin);
  temp[3][3]=0.0;
  PGL_compose (buffer->transform,temp);
  
  PGL_translate_buffer(buffer,PGL_x,1.0);
  PGL_translate_buffer(buffer,PGL_y,-1.0);
  
  PGL_scale_buffer(buffer,PGL_x,(float)buffer->width/2.0);
  PGL_scale_buffer(buffer,PGL_y,-(float)buffer->height/2.0);
  PGL_scale_buffer(buffer,PGL_z,(float)buffer->depth);
  
  buffer->transform_t =1 ;
}


PGL_buffer_t PGL_allocate_buffer (width,height,depth,combiner,format_string)
     unsigned width,height,depth,combiner;
     char* format_string;
{
  unsigned dimensions[3];  
  PGL_buffer_t new = (PGL_buffer_t)malloc(sizeof(PGL_buffer));
  PGL_format_t i;

  new->format = (PGL_format_t)PGL_translate_format (format_string);
  new->height = height;
  new->width = width;
  new->depth = depth;
  new->data_items=0;
  new->data_item_count=0;

  dimensions[0]=PGL_next_power_two(width);
  dimensions[1]=PGL_next_power_two(height);
  dimensions[2]=PGL_next_power_two(depth);

  new->rank=2;
  new->combiner=combiner;
  new->stack_size=1;

  switch(combiner){
  case PGL_queue:
    new->stack_size = depth;
    break;
  case PGL_voxel:
    new->rank=3;
    break;
  }
  
  new->geometry=CM_create_geometry(dimensions,new->rank);
  new->vp_set=CM_allocate_vp_set(new->geometry);  
  new->field = CM_allocate_heap_field_vp_set
    (PGL_point_length(new->format)*new->stack_size,new->vp_set);
  new->preshuffled_t =0;
  new->ambient[0]=new->ambient[1]=new->ambient[2]=255;
  new->lights=(PGL_light_t)0;
  new->light_count=0;
  new->deallocate_t=1;
  PGL_clear_buffer_transform(new);
  PGL_clear_buffer(new,0,0,0);
  return(new);
}

PGL_buffer_t PGL_wrap_buffer_around_field (field,combiner,format_string)
     CM_field_id_t field;
     unsigned combiner;
     char* format_string;
{
  PGL_buffer_t new = (PGL_buffer_t)malloc(sizeof(PGL_buffer));
  PGL_format_t i;

  new->format = (PGL_format_t)PGL_translate_format (format_string);
  new->vp_set = CM_field_vp_set(field);
  new->geometry = CM_vp_set_geometry(new->vp_set);
  new->width = CM_geometry_axis_length(new->geometry,0);
  new->height = CM_geometry_axis_length(new->geometry,1);
  if (CM_geometry_rank(new->geometry) > 2)
    new->depth = CM_geometry_axis_length(new->geometry,2);
  
  new->rank=CM_geometry_rank(new->geometry);
  new->data_items=0;
  new->data_item_count=0;
  new->combiner=combiner;
  new->stack_size=1;
  new->field=field;
  new->preshuffled_t =0;
  new->ambient[0]=new->ambient[1]=new->ambient[2]=255;
  new->lights=(PGL_light_t)0;
  new->light_count=0;
  new->deallocate_t=0;
  PGL_clear_buffer_transform(new);
  return(new);
}

void PGL_deallocate_buffer(buffer)
     PGL_buffer_t buffer;
{
  if (buffer->deallocate_t) {
    CM_deallocate_heap_field(buffer->field);
    CM_deallocate_geometry (buffer->geometry);
    CM_deallocate_vp_set (buffer->vp_set);
  }
  PGL_free_format (buffer->format);
  free (buffer);
}
