#include "pgl_internal.h"
#include <stdio.h>

extern void PGL_open_cmfb_display();
extern void PGL_write_cmfb_display();
extern void PGL_close_cmfb_display();

extern void PGL_open_dv_display();
extern void PGL_write_dv_display();
extern void PGL_close_dv_display();

extern void PGL_open_x_display();
extern void PGL_write_x_display();
extern void PGL_close_x_display();

extern void PGL_open_sgi_display();
extern void PGL_write_sgi_display();
extern void PGL_close_sgi_display();

void PGL_x_windows_nonexistant()
{
  fprintf (stderr,"PGL: X Windows display called without being linked\n");
}

void PGL_cmfs_nonexistant()
{
  fprintf (stderr,"PGL: CMFS display called without being linked\n");
}

void PGL_sgi_nonexistant()
{
  fprintf (stderr,"PGL: SGI display called without being linked\n");
}

static PGL_display_function display_functions[]={
  {PGL_open_cmfb_display,PGL_write_cmfb_display,PGL_close_cmfb_display,0},

#ifdef CMFS
  {PGL_open_dv_display,PGL_write_dv_display,PGL_close_dv_display,0},
#else 
  {PGL_cmfs_nonexistant,PGL_cmfs_nonexistant,PGL_cmfs_nonexistant,0},
#endif

#ifdef X_WINDOWS
  {PGL_open_x_display,PGL_write_x_display,PGL_close_x_display,0},
#else 
  {PGL_x_windows_nonexistant,PGL_x_windows_nonexistant,
     PGL_x_windows_nonexistant,0},
#endif

#ifdef SGI_RENDER
  {PGL_open_sgi_display,PGL_write_sgi_display,PGL_close_sgi_display,0},
#else 
  {PGL_sgi_nonexistant,PGL_sgi_nonexistant,PGL_sgi_nonexistant,0},
#endif

};


PGL_display_t PGL_allocate_display(type,name,width,height)
     unsigned type;
     char *name;
     unsigned width,height;
{
  PGL_display_t rval = (PGL_display_t)malloc(sizeof(PGL_display));
  int temp;
  unsigned dimensions[2];
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;

  rval->type = type;
  rval->width = width;
  rval->height = height;
  (*display_functions[type].open)(rval,name);
  return(rval);
}


void PGL_display_buffer(display,buffer,x,y)
     PGL_display_t display;
     PGL_buffer_t buffer;
     unsigned x,y;
{
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  CM_field_id_t temp;
  CM_geometry_id_t new_geometry;
  CM_vp_set_id_t new_vp_set;
  unsigned dimensions[2];


/* kludge*/ 
  if (buffer->combiner == PGL_voxel){
    dimensions[0]=CM_geometry_axis_length(buffer->geometry,0);
    dimensions[1]=CM_geometry_axis_length(buffer->geometry,1);
    dimensions[1]=MAX(dimensions[1],
		      CM_physical_cube_address_limit/dimensions[0]);
    new_geometry=CM_create_geometry(dimensions,2);
    new_vp_set= CM_allocate_vp_set(new_geometry);
    CM_set_vp_set(new_vp_set);
  } else  CM_set_vp_set(buffer->vp_set);
  temp = CM_allocate_stack_field(24);

  /* this should really be done inside the extracts for device
     independent formats*/
  PGL_extract_buffer(temp,buffer,"ru8 gu8 bu8",
		     display_functions[display->type].preshuffled_t);

  (*display_functions[display->type].write)(display,temp,x,y,buffer);
  CM_deallocate_stack_through(temp);
  CM_set_vp_set(entry_vp_set);
  if (buffer->combiner == PGL_voxel){
    CM_deallocate_geometry(new_geometry);
    CM_deallocate_vp_set(new_vp_set);
  }
}


void PGL_deallocate_display(display)
     PGL_display_t display;
{
  (*display_functions[display->type].close)(display);
/*  CM_deallocate_geometry(CM_vp_set_geometry(display->vp_set));*/
/*  CM_deallocate_vp_set(display->vp_set);*/
  free(display);
}






