#include "pgl_internal.h"

#define EXTRA 1
#define LEN_MULTIPLY 2

unsigned PGL_next_power_two(n)  
     unsigned n;
{
  unsigned i, c = 0, t;
  for (i = 1 << 31; i > 0; i >>= 1)
    if ((i & n) != 0) {
      c++;
      if (c == 1) t = i;
    }
  if (c == 1) return(t);
  else return (t << 1);
}



void PGL_distance (dest,x,y,dt,xt,dl,xl)
     CM_field_id_t dest,x,y;
     unsigned dt,xt;
     unsigned dl,xl;
{
  CM_field_id_t temp = CM_allocate_stack_field (xl);
  CM_field_id_t entry_context;
  
  switch (xt){
  case PGL_float:
    CM_f_subtract_3_1L (temp,x,y,S(xl),E(xl));
    CM_f_f_ceiling_1_1L (dest,S(dl),E(dl));
    CM_f_abs_1_1L (temp,S(xl),E(xl));
    CM_f_add_constant_2_1L (temp,(float)EXTRA,S(xl),E(xl));
    break;
    
  case PGL_signed:
    CM_s_subtract_3_1L (temp,x,y,xl);
    CM_s_abs_1_1L (temp,xl);
    CM_s_add_constant_2_1L (temp,EXTRA,xl);
    break;

  case PGL_unsigned:
    entry_context = CM_allocate_stack_field(1);
    
    CM_store_context(entry_context);
    CM_u_gt_1L (x,y,xl);
    CM_logand_context_with_test();
    CM_u_subtract_3_1L (temp,x,y,xl);
    CM_invert_context();
    CM_logand_context(entry_context);
    CM_u_subtract_3_1L (temp,y,x,xl);
    CM_load_context(entry_context); 
    CM_u_add_constant_2_1L (temp,EXTRA,xl);
    break;
  }
  PGL_cast (dest,temp,dt,xt,dl,xl);
  CM_deallocate_stack_through (temp);
}


void PGL_screen_distance (dest,x1,x2,dt,dl,format)
     CM_field_id_t dest,x1,x2;
     unsigned dt,dl;
     PGL_format_t format;
{
  CM_field_id_t temp = CM_allocate_stack_field(dl);
  
  PGL_distance (temp,PGL_extract(x1,format,PGL_x),
		PGL_extract(x2,format,PGL_x),
		dt,PGL_type(format,PGL_x),
		dl,PGL_length(format,PGL_x));

  PGL_distance (dest,PGL_extract(x1,format,PGL_y),
		PGL_extract(x2,format,PGL_y),
		dt,PGL_type(format,PGL_y),
		dl,PGL_length(format,PGL_y));
  
  switch (dt){
  case PGL_float:
    CM_f_max_2_1L (dest,temp,S(dl),E(dl));
    break;
  case PGL_signed:
    CM_s_max_2_1L (dest,temp,dl);
    break;
  case PGL_unsigned:
    CM_u_max_2_1L (dest,temp,dl);
    break;
  }
  CM_deallocate_stack_through (temp);
}




void PGL_linear_interpolation (dest,s1,s2,xh,total,type,itype,length,ilength)
     CM_field_id_t dest,s1,s2,xh,total;
     unsigned type,itype,length,ilength;
{
  CM_field_id_t temp,temp2, entry_context,swap_context;

  if (type ==  PGL_float){
    temp = CM_allocate_stack_field (length);
    PGL_cast (temp,xh,type,itype,length,ilength);
    CM_f_sub_mult_1L (dest,s2,s1,temp,S(length),E(length));
    PGL_cast (temp,total,type,itype,length,ilength);
    CM_f_divide_2_1L (dest,temp,S(length),E(length));
    CM_f_add_2_1L (dest,s1,S(length),E(length));

/*    printf ("intermediary:");
    PGL_print_range (dest,type,length);
    printf ("\n");*/

  } else {
    temp  = CM_allocate_stack_field (ilength);
    temp2  = CM_allocate_stack_field (ilength);
    PGL_cast (temp2,s1,itype,type,ilength,length);
    PGL_cast (temp,s2,itype,type,ilength,length);
    switch (itype){
    case PGL_float:
      CM_f_sub_mult_1L (temp,temp,temp2,xh,S(ilength),E(ilength));
      CM_f_divide_2_1L (temp,total,S(ilength),E(ilength));
      CM_f_add_2_1L (temp,temp2,S(ilength),E(ilength));
      break;
      
    case PGL_signed:
      CM_s_subtract_2_1L (temp,temp2,ilength);
      CM_s_multiply_2_1L (temp,xh,ilength);
      CM_s_truncate_2_1L (temp,total,ilength);
      CM_s_add_2_1L (temp,temp2,ilength);
      break;
    
    case PGL_unsigned:
      printf ("error from hell.\n");
    }
    PGL_cast (dest,temp,type,itype,length,ilength);
  }
  CM_deallocate_stack_through (temp);
}


void PGL_interpolate_field(dest,s1,s2,format,t,total,ptype,plength)
     CM_field_id_t dest,s1,s2;
     PGL_format_t format;
     CM_field_id_t t,total;
     unsigned ptype,plength;
{
  PGL_format_t i;
  for (i=format;i!=(PGL_format_t)0;i=i->next)
    if (!i->constant_t)
      PGL_linear_interpolation (PGL_extract(dest,format,i->name),
				PGL_extract(s1,format,i->name),
				PGL_extract(s2,format,i->name),
				t,total,
				i->type,ptype,
				i->length,plength);
}

