#include "pgl_internal.h"
#include "contours.h"

extern PGL_fill_triangle();
extern PGL_fill_line();

static CM_field_id_t table=0;

void PGL_fill_sgi_triangle(frame)
     PGL_spawn_frame_t frame;
{
  PGL_write_sgi (frame->dest,frame->xp,frame->format,frame->size,
		frame->flags);
}


void PGL_allocate_stack_field_vector(sdest,element_length,vector_length)
     CM_field_id_t *sdest;
     unsigned element_length,vector_length;
{
  int i;
  sdest[0]=CM_allocate_stack_field(element_length*vector_length);
  for (i=1;i<vector_length;i++)
    sdest[i]=CM_add_offset_to_field_id(sdest[i-1],element_length);
}

void PGL_load_iso_table()
{
  int i,j,k;
  CM_vp_set_id_t entry_vp_set=CM_current_vp_set;
  CM_field_id_t temp,index,entry_context;

  CM_set_vp_set(CM_physical_vp_set());
  temp = CM_allocate_stack_field(64);
  index = CM_allocate_stack_field(32);
  table=CM_allocate_heap_field(256);
  entry_context=CM_allocate_stack_field(1);
  
  CM_store_context(entry_context);
  CM_set_context();
  for(i=0;i<128;i++){
    for(j=0;j<4;j++)
      for(k=0;k<3;k++)
	CM_u_move_const_always_1L(temp+(j*3+k)*4,contours[i*12+j*3+k],4);
    CM_u_move_const_always_1L (index,i,32);
    CM_aset32_shared_2L (temp,table,index,64,7,128);
  }
  CM_load_context(entry_context);
  CM_deallocate_stack_through(temp);
  CM_set_vp_set(entry_vp_set);
}



void PGL_3d_field_normals(dest,source,d_format,s_format,flags)
{
  CM_field_id_t t=CM_allocate_stack_field(32);
  CM_field_id_t pvalue=CM_allocate_stack_field(32);
  int axes[3];
  
  if (flags & PGL_fortran_geometry){
    axes[0]=1;axes[1]=2;axes[2]=3;
  } else {
    axes[0]=0;axes[1]=1;axes[2]=2;
  }
  PGL_extract_copy(pvalue,source,s_format,PGL_v,PGL_float,32);

  CM_get_from_news_always_1L(t,pvalue,axes[0],CM_upward,32);
  CM_f_news_sub_3_1L (PGL_extract(dest,d_format,PGL_X),t,pvalue,
		      axes[0],CM_downward,23,8);
  CM_get_from_news_always_1L(t,pvalue,axes[1],CM_upward,32);
  CM_f_news_sub_3_1L (PGL_extract(dest,d_format,PGL_Y),t,pvalue,
		      axes[1],CM_downward,23,8);
  CM_get_from_news_always_1L(t,pvalue,axes[2],CM_upward,32);
  CM_f_news_sub_3_1L (PGL_extract(dest,d_format,PGL_Z),t,pvalue,
		      axes[2],CM_downward,23,8);

  CM_deallocate_stack_through(t);
}



void PGL_write_isosurface(frame,value,format,point)
     PGL_spawn_frame_t frame;
     float value;
     PGL_format_t format;
     CM_field_id_t point;
{
  int i,j,k,axis,point_length,axes[3],axes2[3];
  CM_field_id_t state=CM_allocate_stack_field(8);
  CM_field_id_t t=CM_allocate_stack_field(32);
  CM_field_id_t t2=CM_allocate_stack_field(32);
  CM_field_id_t entry_context=CM_allocate_stack_field(1);
  CM_field_id_t pvalue=CM_allocate_stack_field(32);
  CM_field_id_t triangle[3],edges[12],test[12];
  CM_field_id_t desc=CM_allocate_stack_field(64);
  CM_field_id_t go=CM_allocate_stack_field(1);

  
  if (frame->flags & PGL_fortran_geometry){
    axes[0]=1;axes[1]=2;axes[2]=3;
    axes2[0]=1;axes2[1]=3;axes2[2]=2;
  } else {
    axes[0]=0;axes[1]=1;axes[2]=2;
    axes2[0]=0;axes2[1]=2;axes2[2]=1;
  }

  point_length = PGL_point_length(frame->format);

  CM_store_context(entry_context);
  CM_set_context();
  PGL_allocate_stack_field_vector(triangle,point_length,3);
  PGL_allocate_stack_field_vector(edges,point_length,12);

  PGL_generate_transfer_field(triangle[0],point,frame->dest,
			      frame->format,format,frame->flags,
			      PGL_3d_field_normals);
  
  PGL_extract_copy(pvalue,point,format,PGL_v,PGL_float,32);

  CM_f_lt_constant_1L (pvalue,value,23,8);
  CM_store_test(state);
  CM_get_from_news_always_1L (state+1,state,axes2[0],CM_upward,1);
  CM_get_from_news_always_1L (state+2,state,axes2[1],CM_upward,2);
  CM_get_from_news_always_1L (state+4,state,axes2[2],CM_upward,4);

  for (i=0;i<7;i++) CM_logxor_2_1L (state+i,state+7,1);
  
  CM_aref32_shared_2L (desc,table,state,64,7,128);
  
  CM_f_subfrom_constant_3_1L (t,pvalue,value,23,8);
  for (axis=0;axis<3;axis++){
    CM_get_from_news_always_1L (t2,pvalue,axes[axis],CM_upward,32);
    CM_f_subtract_always_2_1L (t2,pvalue,23,8);
    CM_get_from_news_always_1L (triangle[2],triangle[0],axes[axis],
				CM_upward,point_length);
    
    PGL_interpolate_field(edges[axis],triangle[0],triangle[2],frame->format,
			  t,t2,PGL_float,32);
  }

  for (j=0;j<3;j++)
    for (i=0;i<2;i++){
      CM_get_from_news_always_1L (edges[3+j*2+i],edges[(i+j+1)%3],axes[j],
				  CM_upward,point_length);
    }
  
  for (i=0;i<3;i++){
    CM_get_from_news_always_1L (edges[9+i],edges[((i+1)%3)*2+3],axes[i],
				CM_upward,point_length);
  }
  
  for (i=0;i<4;i++){
    CM_u_move_zero_always_1L (go,1);
    for (j=0;j<3;j++)
      for (k=0;k<12;k++)
	if (move_map[i*3+j][k]){
	  CM_u_eq_constant_1L (desc+i*12+4*j,k+1,4);
	  CM_logand_context_with_test();
	  CM_u_move_constant_1L(go,1,1);
	  CM_u_move_1L (triangle[j],edges[k],point_length);
	  CM_set_context();
	}
    CM_load_context(go);
    CM_logand_context(entry_context);
    if (frame->dest->combiner != PGL_sgi_render)
      PGL_plot_triangle_internal(frame,frame->dest,frame->format,triangle[0]);
    else {
      CM_u_move_constant_1L (t2,1,32);
      PGL_write_spawn (frame,t2,triangle[0]);
    }
    CM_set_context();
  }
  CM_load_context(entry_context);
  CM_deallocate_stack_through(state);
}



PGL_spawn_frame_t PGL_open_isosurface (dest,format,size1,size2,flags)
     PGL_buffer_t dest;
     PGL_format_t format;
     unsigned size1,size2;
     unsigned flags;
{
  PGL_format_t i_format=PGL_generate_transfer_format(dest,format,flags);
  unsigned point_length = PGL_point_length(i_format);
  PGL_spawn_frame_t frame;

  if (table==0) PGL_load_iso_table();
  if (dest->combiner == PGL_sgi_render)
    frame = PGL_open_spawn(dest,PGL_fill_sgi_triangle,i_format,8192,
			   3*point_length,flags);
    
  else
    frame = (PGL_spawn_frame_t)
      PGL_open_triangle (dest,i_format,size1,size2,flags);
  return (frame);
}


void PGL_close_isosurface(frame)
     PGL_spawn_frame_t frame;
{

  if (frame->dest->combiner == PGL_sgi_render)
    PGL_close_spawn(frame);    
  else 
    PGL_close_triangle(frame);
}

void PGL_plot_isosurface(dest,value,flags,format_string,point)
     PGL_buffer_t dest;
     float value;
     unsigned flags;
     char *format_string;
     CM_field_id_t point;
{
  PGL_format_t format=PGL_translate_format(format_string);
  PGL_spawn_frame_t frame=PGL_open_isosurface(dest,format,65536,262144,flags);
  PGL_write_isosurface(frame,value,format,point);
  PGL_close_isosurface(frame);
}
