#include "pgl_internal.h"
#include <math.h>
#include <stdio.h>

void PGL_apply_lighting(dest,c_source,n_source,buffer,
			d_format,c_format,n_format,length)
     CM_field_id_t dest,c_source,n_source;
     PGL_buffer_t buffer;
     PGL_format_t d_format,c_format,n_format;
     unsigned length;
{
  CM_field_id_t temp = CM_allocate_stack_field(length);
  CM_field_id_t dot = CM_allocate_stack_field(length);
  CM_field_id_t norm = CM_allocate_stack_field(length);
  CM_field_id_t color = CM_allocate_stack_field(3*length);
  PGL_light_t light;

  PGL_extract_copy(temp,c_source,c_format,PGL_r,PGL_float,length);
  CM_f_multiply_constant_3_1L (color,temp,(float)buffer->ambient[0]/255.0,
			       S(length),E(length));

  PGL_extract_copy(temp,c_source,c_format,PGL_g,PGL_float,length);
  CM_f_multiply_constant_3_1L (CM_add_offset_to_field_id(color,length),
			     temp,(float)buffer->ambient[1]/255.0,
			     S(length),E(length));

  PGL_extract_copy(temp,c_source,c_format,PGL_b,PGL_float,length);
  CM_f_multiply_constant_3_1L (CM_add_offset_to_field_id(color,2*length),
			     temp,(float)buffer->ambient[2]/255.0,
			     S(length),E(length));

  /* should a component of the normal not exist, it should be
     set to zero....oh defaults!*/
  for (light=buffer->lights;light!=(PGL_light_t)0;light=light->next){

    /* these extract copies with normalization could be done one
       to increase speed for multiple light sources with only
       'length' more bits*/
    PGL_extract_copy (temp,n_source,n_format,PGL_X,PGL_float,length);
    CM_f_multiply_constant_3_1L (dot,temp,light->position[0],
				 S(length),E(length));
    CM_f_multiply_3_1L (norm,temp,temp,S(length),E(length));
    
    PGL_extract_copy (temp,n_source,n_format,PGL_Y,PGL_float,length);
    CM_f_mult_const_add_1L(dot,temp,light->position[1],dot,
			   S(length),E(length));
    CM_f_mult_add_1L (norm,temp,temp,norm,S(length),E(length));
    
    PGL_extract_copy (temp,n_source,n_format,PGL_Z,PGL_float,length);
    CM_f_mult_const_add_1L(dot,temp,light->position[2],dot,
			   S(length),E(length));
    CM_f_mult_add_1L (norm,temp,temp,norm,S(length),E(length));
    
    CM_f_sqrt_1_1L (norm,S(length),E(length));
    CM_f_divide_2_1L (dot,norm,S(length),E(length));
    CM_f_multiply_constant_2_1L(dot,1.0/sqrt
				(light->position[0]*light->position[0]+
				 light->position[1]*light->position[1]+
				 light->position[2]*light->position[2]),
				S(length),E(length));

    if (PGL_find_name (c_format,PGL_s)){
      PGL_extract_float_range (temp,c_source,c_format,PGL_s,32);
      CM_f_f_power_2_1L (dot,temp,23,8);
    }
	
	CM_f_max_constant_2_1L (dot,0.0,S(length),E(length));

    PGL_extract_copy(temp,c_source,c_format,PGL_r,PGL_float,length);
    CM_f_multiply_2_1L (temp,dot,S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,1.0-(float)buffer->ambient[0]/255.0,
				 S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,(float)light->color[0]/255.0,
				S(length),E(length));
    CM_f_add_2_1L (color,temp,S(length),E(length));


    PGL_extract_copy(temp,c_source,c_format,PGL_g,PGL_float,length);
    CM_f_multiply_2_1L (temp,dot,S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,1.0-(float)buffer->ambient[1]/255.0,
				 S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,(float)light->color[1]/255.0,
				S(length),E(length));
    CM_f_add_2_1L (CM_add_offset_to_field_id(color,length),
			    temp,S(length),E(length));
    
    
    PGL_extract_copy(temp,c_source,c_format,PGL_b,PGL_float,length);
    CM_f_multiply_2_1L (temp,dot,S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,1.0-(float)buffer->ambient[2]/255.0,
				 S(length),E(length));
    CM_f_multiply_constant_2_1L (temp,(float)light->color[2]/255.0,
				 S(length),E(length));
    CM_f_add_2_1L (CM_add_offset_to_field_id(color,2*length),
		   temp,S(length),E(length));

  }
  PGL_cast(PGL_extract(dest,d_format,PGL_r),color,
	   PGL_type(d_format,PGL_r),PGL_float,
	   PGL_length(d_format,PGL_r),length);
  
  PGL_cast(PGL_extract(dest,d_format,PGL_g),
	   CM_add_offset_to_field_id(color,length),
	   PGL_type(d_format,PGL_g),PGL_float,
	   PGL_length(d_format,PGL_g),length);
  
  PGL_cast(PGL_extract(dest,d_format,PGL_b),
	   CM_add_offset_to_field_id(color,2*length),
	   PGL_type(d_format,PGL_b),PGL_float,
	   PGL_length(d_format,PGL_b),length);

  CM_deallocate_stack_through(temp);
}

int PGL_add_light (buffer,x,y,z,r,g,b)
     PGL_buffer_t buffer;
     float x,y,z;
     unsigned r,g,b;
{
 PGL_light_t new;
 int temp;
 
 if (buffer->combiner == PGL_sgi_render){
   float ambient[3];
   float color[3];
   float position[3];
   color[0]=(float)r/255.0;
   color[1]=(float)g/255.0;
   color[2]=(float)b/255.0;
   position[0]=x;
   position[1]=y;
   position[2]=z;
   ambient[0]=(float)(buffer->ambient[0])/255.0;
   ambient[1]=(float)(buffer->ambient[1])/255.0;
   ambient[1]=(float)(buffer->ambient[2])/255.0;
   PGL_sgi_initialize();
   temp = SGI_add_light(ambient,color,position);
   SGI_light_on(temp);
   return (temp);

 } else {
   new = (PGL_light_t)malloc(sizeof(PGL_light));
   new->position[0]=x;
   new->position[1]=y;
   new->position[2]=z;
   new->color[0]=r;
   new->color[1]=g;
   new->color[2]=b;
   new->next = buffer->lights;
   buffer->lights = new;
   return(new->light_number=(++buffer->light_count));
 }
}


void PGL_delete_light (buffer,light_no)
     PGL_buffer_t buffer;
     unsigned light_no;
{
  PGL_light_t *i,t;

  if (buffer->combiner == PGL_sgi_render) SGI_light_off (light_no);
  else
    for (i=(&buffer->lights);*i;i=(&((*i)->next)))
      if ((*i)->light_number != light_no){
	t=(*i);
	(*i)->next=(*i)->next->next;
	free(t);
      }
}
