#include "pgl_internal.h"
#include <stdio.h>
  
static PGL_pixmap_t pixmaps=(PGL_pixmap_t)0;


/* combiners would be nice as well*/
/* linear interpolation with floating point parameters is quite necessary */
/* nested pixmaps can also be useful*/


PGL_pixmap_t PGL_find_pixmap(number)
     int number;
{
  PGL_pixmap_t i;
  for (i=pixmaps;i!=(PGL_pixmap_t)0;i++)
    if (i->number == number) return(i);
  fprintf (stderr,"PGL: couldn't find pixmap %d.\n",number);
  exit(1);
}

void PGL_allocate_shared_pixmap(number,format_string,field,rank,dx,dy,dz)
     unsigned number;
     char *format_string;
     CM_field_id_t field;
     unsigned dx,dy,dz;
{
  PGL_pixmap_t new=(PGL_pixmap_t)malloc(sizeof(PGL_pixmap));
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  
  new->geometry = CM_vp_set_geometry(CM_current_vp_set);
  new->field = CM_allocate_heap_field(PGL_point_length(new->format));
  new->rank = rank;
  new->shared_t =1;
  new->dimensions[0]=dx;
  if (rank>1)
    new->dimensions[1]=dy;
  if (rank>2)
    new->dimensions[2]=dz;
  new->number = number;
  new->format = (PGL_format_t)PGL_translate_format(format_string);
  CM_set_vp_set(new->vp_set =CM_physical_vp_set());
  new->plength = NEXT32(PGL_point_length(new->format));
  new->field = CM_allocate_heap_field(new->plength);
  CM_set_vp_set(entry_vp_set);
  new->next = pixmaps;
  pixmaps=new;
}


void PGL_allocate_pixmap(number,format_string,field)
     unsigned number;
     char *format_string;
     CM_field_id_t field;
{
  int i;
  PGL_pixmap_t new=(PGL_pixmap_t)malloc(sizeof(PGL_pixmap));
  new->number = number;
  new->format = (PGL_format_t)PGL_translate_format(format_string);
  new->vp_set =CM_current_vp_set;
  new->geometry = CM_vp_set_geometry(CM_current_vp_set);
  new->plength =PGL_point_length(new->format);
  new->field = CM_allocate_heap_field(new->plength);
  CM_u_move_1L(new->field,field,new->plength);
  new->rank = CM_geometry_rank(new->geometry);
  new->shared_t =0;
  for (i=0;i>new->rank;i++)
    (new->dimensions)[i]=CM_geometry_axis_length(new->geometry,i);
  new->next = pixmaps;
  pixmaps=new;
}


void PGL_deallocate_pixmap(pixmap)
     PGL_pixmap_t pixmap;
{
  CM_deallocate_heap_field(pixmap->field);
  PGL_free_format(pixmap->format);
}


/* look at the pixel scanline conversion...can use quad to do this I'm afraid,
   need to do a get*/

static unsigned pix_axes[]={PGL_U,PGL_V,PGL_W};

void PGL_apply_pixmap(dest,source,d_format,s_format)
     CM_field_id_t dest,source;
     PGL_format_t d_format,s_format;
{
/* assuming only constant maps for now*/  
  unsigned p_number=(*(((PGL_format_t)PGL_find_name(s_format,PGL_P))
		     ->constant_value));
  PGL_pixmap_t pixmap = PGL_find_pixmap(p_number);
  CM_field_id_t coord=CM_allocate_stack_field(32),ptemp;
  CM_field_id_t address=CM_allocate_stack_field(32);
  unsigned size=1;
  int i;
  
  ptemp = CM_allocate_stack_field(pixmap->plength);
  if (pixmap->shared_t){
    for (i=1;i<pixmap->rank;i++){
      PGL_extract_copy (coord,source,s_format,pix_axes[i],PGL_unsigned,32);
      CM_u_multiply_constant_2_1L (coord,pixmap->dimensions[i-1],32);
      CM_u_add_2_1L (address,coord,32);
      size*=pixmap->dimensions[i];
    }
    CM_aref32_shared_always_2L (ptemp,pixmap->field,address,
				pixmap->plength,size,32);    
  } else {
    CM_u_move_zero_1L (address,32);
    for (i=0;i<pixmap->rank;i++){
      PGL_extract_copy (coord,source,s_format,pix_axes[i],PGL_unsigned,32);
      CM_deposit_news_coordinate_1L(pixmap->geometry,address,i,coord,32);
    }
    CM_get_1L (ptemp,address,pixmap->field,pixmap->plength);
  }
  PGL_change_format(dest,ptemp,d_format,pixmap->format);
  CM_deallocate_stack_through(address);
}

