#include "pgl_internal.h"
#include <stdio.h>
#include <stdio.h>

#define EXTRA 4


PGL_format_t PGL_generate_transfer_format (dest,source,flags)
     PGL_buffer_t dest;
     PGL_format_t source;
     unsigned flags;
{
  PGL_format_t format;
  
  unsigned xl=CM_geometry_coordinate_length(dest->geometry,0)+EXTRA;
  unsigned yl=CM_geometry_coordinate_length(dest->geometry,1)+EXTRA;
  unsigned zl;
  
  format=(PGL_format_t)PGL_copy_format(source);

  if (dest->combiner != PGL_sgi_render)
    if (PGL_find_name (dest->format,PGL_z)){
      zl=PGL_length(dest->format,PGL_z)+EXTRA;
      PGL_append_format (&format,PGL_z,PGL_signed,zl,0,0);
    }
  
  if ((flags & PGL_generate_normals) && (flags & PGL_interpolate_normals)) {
    PGL_push_format(format,PGL_X,PGL_float,32,0,0);
    PGL_push_format(format,PGL_Y,PGL_float,32,0,0);
    PGL_push_format(format,PGL_Z,PGL_float,32,0,0);
  }
  if (dest->lights){
    PGL_push_format (&format,PGL_r,
		     PGL_type(dest->format,PGL_r),
		     PGL_length(dest->format,PGL_r),0,0);
    PGL_push_format (&format,PGL_g,
		     PGL_type(dest->format,PGL_g),
		     PGL_length(dest->format,PGL_g),0,0);
    PGL_push_format (&format,PGL_b,
		     PGL_type(dest->format,PGL_b),
		     PGL_length(dest->format,PGL_b),0,0);
  }
  
  /* for use by contour and isocontour*/
  PGL_delete_format (&format,PGL_v);
  
  PGL_append_format (&format,PGL_x,PGL_signed,xl,0,0);
  PGL_append_format (&format,PGL_y,PGL_signed,yl,0,0);
  if (dest->combiner == PGL_voxel){
    zl=CM_geometry_coordinate_length(dest->geometry,2)+EXTRA;
    PGL_append_format (&format,PGL_z,PGL_signed,zl,0,0);
  }
  return (format);
}



void PGL_generate_transfer_field (dest,source,buffer,dest_format,
				  source_format,flags,normal_function)
     CM_field_id_t dest,source;
     PGL_buffer_t buffer;
     PGL_format_t dest_format,source_format;
     unsigned flags;
     function_pointer normal_function;

{
  CM_field_id_t temp;
  CM_field_id_t ndest;
  PGL_format_t n_format;
  int allocate_t=0;

  if (flags & PGL_print_inputs) PGL_print_field(source,source_format);

  PGL_change_format (dest,source,dest_format,source_format);
  
  if (buffer->transform_t){
    temp = CM_allocate_stack_field(3*32);
    
    PGL_extract_copy (temp,source,source_format,PGL_x,PGL_float,32);
    PGL_extract_copy (CM_add_offset_to_field_id(temp,32),source,source_format,
		      PGL_y,PGL_float,32);

    if (PGL_find_name (source_format,PGL_z))
      PGL_extract_copy (CM_add_offset_to_field_id(temp,64),source,
			source_format,PGL_z,PGL_float,32);
    else CM_f_move_zero_1L (CM_add_offset_to_field_id(temp,64),23,8);

    PGL_transform (temp,temp,buffer->transform,32);

    PGL_cast (PGL_extract(dest,dest_format,PGL_x),temp,
	      PGL_type(dest_format,PGL_x),PGL_float,
	      PGL_length(dest_format,PGL_x),32);

    PGL_cast (PGL_extract(dest,dest_format,PGL_y),
	      CM_add_offset_to_field_id(temp,32),
	      PGL_type(dest_format,PGL_y),PGL_float,
	      PGL_length(dest_format,PGL_y),32);

    if (PGL_find_name(dest_format,PGL_z))
      PGL_cast (PGL_extract(dest,dest_format,PGL_z),
		CM_add_offset_to_field_id(temp,64),
		PGL_type(dest_format,PGL_z),PGL_float,
		PGL_length(dest_format,PGL_z),32);
  }

  if (normal_function && (flags & PGL_generate_normals)) {
    if (flags & PGL_interpolate_normals){
      ndest=dest;
      n_format=dest_format;
    } else {
      ndest = CM_allocate_stack_field(96);
      n_format = PGL_translate_format("Xf Yf Zf");
      allocate_t=1;
    }
    (*normal_function) (ndest,source,n_format,source_format,flags);
    if (flags & PGL_buffer_normalization){
      CM_f_divide_constant_2_1L (PGL_extract(ndest,n_format,PGL_X),
				 (float)buffer->width,23,8);
      CM_f_divide_constant_2_1L (PGL_extract(ndest,n_format,PGL_Y),
				 (float)buffer->height,23,8);
      CM_f_divide_constant_2_1L (PGL_extract(ndest,n_format,PGL_Z),
				 (float)buffer->depth,23,8);
    }
  } else {
    ndest=source;
    n_format=source_format;
  }
  if ((!(flags & PGL_interpolate_normals)) && (buffer->lights)
      && (! ((flags&PGL_generate_normals) && (!normal_function)))) {
    PGL_apply_lighting(dest,source,ndest,buffer,
		       dest_format,source_format,n_format,32);
  }
  if (flags & PGL_print_generated_normals){
    PGL_print_field (ndest,n_format);
  }
  if (allocate_t){
    CM_deallocate_stack_through (ndest);
    PGL_free_format (n_format);
  }
  if (flags & PGL_print_transformed_inputs)
    PGL_print_field(dest,dest_format);
}



void PGL_plot_point(dest,flags,format_string,point)
     PGL_buffer_t dest;
     unsigned flags;
     char *format_string;
     CM_field_id_t point;
{
  PGL_format_t format=(PGL_format_t)PGL_translate_format(format_string);
  PGL_format_t i_format=PGL_generate_transfer_format (dest,format,flags);
  unsigned point_length= PGL_point_length (i_format);
  CM_field_id_t args=CM_allocate_stack_field (point_length);
  
  PGL_generate_transfer_field(args,point,dest,i_format,format,flags,0);
  PGL_send(dest,args,i_format,flags);
  CM_deallocate_stack_through (args);
  PGL_free_format (i_format);
  PGL_free_format (format);
}
  

