#include "pgl_internal.h"

void PGL_overwrite_combiner(dest,address,source)
     PGL_buffer_t dest;
     unsigned address;
     CM_field_id_t source;
{
  CM_send_with_overwrite_1L (dest->field,address,source,
			     PGL_point_length(dest->format),CM_no_field);  
}


void PGL_max_combiner(dest,address,source)
     PGL_buffer_t dest;
     unsigned address;
     CM_field_id_t source;
{
  CM_send_with_u_max_1L (dest->field,address,source,
			 PGL_point_length(dest->format),CM_no_field);    
}

void PGL_queue_combiner(dest,address,source)
     PGL_buffer_t dest;
     unsigned address;
     CM_field_id_t source;
{
  CM_send_to_queue32_1L (dest->field,address,source,32,
			 NEXT32(dest->stack_size));
}



void PGL_add_combiner(dest,address,source)
     PGL_buffer_t dest;
     unsigned address;
     CM_field_id_t source;
{
  PGL_format_t i;
  CM_vp_set_id_t entry_vp_set;
  for (i=dest->format;i!=(PGL_format_t)0;i=i->next)
    if (!i->constant_t)
      switch (i->type){
      case PGL_unsigned:
	CM_send_with_u_add_1L  (PGL_extract(dest->field,dest->format,i->name),
				address,
				PGL_extract(source,dest->format,i->name),
				i->length,CM_no_field);  
	break;
	
      case PGL_signed:
	CM_send_with_s_add_1L  (PGL_extract(dest->field,dest->format,i->name),
				address,
				PGL_extract(source,dest->format,i->name),
				i->length,CM_no_field);  
	break;
	
      case PGL_float:
	CM_send_with_f_add_1L  (PGL_extract(dest->field,dest->format,i->name),
				address,
				PGL_extract(source,dest->format,i->name),
				S(i->length),E(i->length),CM_no_field);  
	break;
      }
}


static function_pointer combiners[]={
  PGL_overwrite_combiner,
  PGL_max_combiner,
  PGL_queue_combiner,
  PGL_add_combiner,
  PGL_overwrite_combiner   /*voxel*/
};


void PGL_send_address (dest,buffer,format,source)
     CM_field_id_t dest;
     PGL_buffer_t buffer;
     PGL_format_t format;
     CM_field_id_t source;
{
  unsigned xl=CM_geometry_coordinate_length(buffer->geometry,0);
  unsigned yl=CM_geometry_coordinate_length(buffer->geometry,1);
  unsigned zl;
  CM_field_id_t x = CM_allocate_stack_field(xl);
  CM_field_id_t y = CM_allocate_stack_field(yl);
  CM_field_id_t z;
  

  PGL_extract_copy (x,source,format,PGL_x,PGL_unsigned,xl);
  PGL_extract_copy (y,source,format,PGL_y,PGL_unsigned,yl);

  if (buffer->preshuffled_t)
    CMFB_shuffle_from_x_y (dest,x,y,buffer->geometry);
  else{
    CM_make_news_coordinate_1L (buffer->geometry,dest,0,x,xl);
    CM_deposit_news_coordinate_1L (buffer->geometry,dest,1,y,yl);
  }

  if (buffer->combiner == PGL_voxel){
    zl=CM_geometry_coordinate_length(buffer->geometry,2);
    z=CM_allocate_stack_field(zl);
    PGL_extract_copy (z,source,format,PGL_z,PGL_unsigned,zl);
    CM_deposit_news_coordinate_1L (buffer->geometry,dest,2,z,zl);
  }
  CM_deallocate_stack_through(x);
}


/* as usual, total reorganization is required */

void PGL_send (dest,source,format,flags)
     PGL_buffer_t dest;
     CM_field_id_t source;
     PGL_format_t format;
     unsigned flags;
{
  CM_field_id_t address = CM_allocate_stack_field(32);
  CM_field_id_t assembled=CM_allocate_stack_field
    (PGL_point_length(dest->format));
  CM_field_id_t entry_context = CM_allocate_stack_field(1);
  PGL_format_t i,j;
  CM_vp_set_id_t entry_vp_set= CM_current_vp_set;

  if (flags & PGL_print_send_input)  PGL_print_field(source,format);

  if (PGL_find_name (format,PGL_R))
    PGL_send_sphere (dest,source,format);
  else{ 
    CM_store_context(entry_context);
    PGL_clip_point(CM_context_flag,dest,format,source);
    
    PGL_change_format (assembled,source,dest->format,format);
    
    /*  PGL_print_field(source,format);*/
    if (PGL_find_name (format,PGL_P))
      PGL_apply_pixmap (assembled,source,dest->format,format);
    
    if ((flags & PGL_interpolate_normals) && (dest->lights))
      PGL_apply_lighting(assembled,assembled,source,dest,
			 dest->format,dest->format,format,32);
    PGL_send_address (address,dest,format,source);

    if (flags & PGL_print_assembled_input)  
      PGL_print_field(assembled,dest->format);

    (*(combiners[dest->combiner]))(dest,address,assembled);
    
    CM_load_context (entry_context);
    CM_deallocate_stack_through (address);
  }
  if (flags & PGL_print_buffer_after_send){
    CM_set_vp_set(dest->vp_set);
    PGL_print_field(dest->field,dest->format);
    CM_set_vp_set(entry_vp_set);
  }
}



