#include <sgi.h>
#include <pgl_internal.h>
#include <cm/cmfs.h>
#include <cm/cm_file.h>

#define HOSTNAME "athena"

#define VERTEX_LENGTH 96
#define NORMAL_LENGTH 96
#define COLOR_LENGTH 24
#define COLOR_PADDING 0
  
static int initialized=0;
static int fd;

float from[3]={700.0,700.0,700.0};
float to[3]={0.0,0.0,0.0};

static float zero[]={0,0,0};
static float one[]={1,1,1};

void PGL_sgi_initialize()
{
  if (!initialized){
    fd=SGI_open (HOSTNAME);
    initialized++;
    SGI_view (from,to,0.0);
  }
}



void PGL_sgi_add_light(x,y,z,r,g,b)
     float x,y,z;
     unsigned char r,g,b;
{
  float light_color[3];
  float light_from[3];
 
  light_color[0]=r/255.0;
  light_color[1]=g/255.0;
  light_color[2]=b/255.0;
 
  light_from[0]=x;
  light_from[1]=y;
  light_from[2]=z;

  if (!initialized) PGL_sgi_initialize();    
  SGI_light_on(SGI_add_light (zero,light_color,light_from));
}


void PGL_sgi_ambient(r,g,b)
     unsigned r,g,b;
{
  float ambient[3];
  
  ambient[0]=r/255.0;
  ambient[1]=g/255.0;
  ambient[2]=b/255.0;
  
  if (!initialized) PGL_sgi_initialize();    
  SGI_light_on(SGI_add_light (ambient,zero,one));
}



void PGL_sgi_extract()
{
}


void PGL_open_sgi_display()
{
}


/* these fields should really be part of the display*/
void PGL_write_sgi_display(display,field,x,y,buffer)
     PGL_display_t display;
     unsigned x,y;
     PGL_buffer_t buffer;
{
  if (buffer->old_data_item_count >= 0){
    SGI_deactivate_list(buffer->old_data_item_count,
			buffer->old_data_items);
    SGI_delete_list(buffer->old_data_item_count,
		    buffer->old_data_items);
    free (buffer->old_data_items);
    buffer->old_data_items=0;
    buffer->old_data_item_count=0;
  }
  SGI_activate_list(buffer->data_item_count,buffer->data_items);
/*  SGI_rot (3.0,5.0,8.0);*/
  SGI_add_view (from,to,0.0);
}


void PGL_close_sgi_display()
{
}


void PGL_write_sgi(buffer,fields,format,size,flags)
     PGL_buffer_t buffer;
     CM_field_id_t fields;
     PGL_format_t format;
     unsigned size,flags;
{
  unsigned point_length = PGL_point_length(format);
  unsigned number=size/point_length;
  unsigned sgi_point_length = 96;
  unsigned sgi_flags=0;
  CM_field_id_t temp,index,foo;
  int i;
  
  if (!initialized) PGL_sgi_initialize(flags);    

  if (flags & PGL_sgi_server_debug){
    SGI_debug(1);
/*    SGI_vertex_debug(1);*/
  }


  if (PGL_find_name (format,PGL_X)||PGL_find_name (format,PGL_Y)
      ||PGL_find_name (format,PGL_Z)){
    sgi_flags |= SGI_NORMALS;
    sgi_point_length += 96;
  }

  
  if (PGL_find_name (format,PGL_r)|| PGL_find_name (format,PGL_g) || 
      PGL_find_name (format,PGL_b)){
    sgi_flags |= SGI_COLORS;
    sgi_point_length += 24;
  }

  index=temp = CM_allocate_stack_field(sgi_point_length*number);
  foo = CM_allocate_stack_field(32);
  for (i=0;i<number;i++) PGL_print_field(fields+i*point_length,format);
  for (i=0;i<number;i++){
    PGL_extract_copy(index,fields+i*point_length,format,PGL_x,PGL_float,32);
    index+=32;
    PGL_extract_copy(index,fields+i*point_length,format,PGL_y,PGL_float,32);
    index+=32;
    PGL_extract_copy(index,fields+i*point_length,format,PGL_z,PGL_float,32);
    index+=32;
  }

  if (sgi_flags & SGI_NORMALS){  
    for (i=0;i<number;i++){
      PGL_extract_copy(index,fields+i*point_length,format,PGL_X,
			  PGL_float,32);
      CM_f_multiply_3_1L (foo,index,index,23,8);
      
      index+=32;
      PGL_extract_copy(index,fields+i*point_length,format,PGL_Y,
		       PGL_float,32);
      CM_f_mult_add_1L (foo,index,index,foo,23,8);
      index+=32;
      PGL_extract_copy(index,fields+i*point_length,format,PGL_Z,
		       PGL_float,32);
      CM_f_mult_add_1L (foo,index,index,foo,23,8);
      CM_f_sqrt_1_1L (foo,23,8);
      CM_f_divide_2_1L (index,foo,23,8);
      CM_f_divide_2_1L (index-32,foo,23,8);
      CM_f_divide_2_1L (index-64,foo,23,8);
      index+=32;
    }
  }

  if (sgi_flags & SGI_COLORS){  
    for (i=0;i<number;i++){
      PGL_extract_copy(index,fields+i*point_length,format,PGL_r,
			  PGL_unsigned,8);
      index+=8;
      PGL_extract_copy(index,fields+i*point_length,format,PGL_g,
			  PGL_unsigned,8);
      index+=8;
      PGL_extract_copy(index,fields+i*point_length,format,PGL_b,
			  PGL_unsigned,8);
      index+=8;
    }
  }
/*

  if (flags&PGL_clip_news_edges){
    CM_field_id_t context=CM_allocate_stack_field(1);
    CM_field_id_t address=CM_allocate_stack_field(32);
    CM_geometry_id_t entry_geometry = CM_vp_set_geometry(CM_current_vp_set);
    
    CM_store_context(context);
    for (i=0;i<CM_geometry_rank(entry_geometry);i++){
      CM_my_news_coordinate_1L (address,i,32);
      CM_u_lt_constant_1L (address,
			   CM_geometry_axis_length(entry_geometry,i)-1,32);
      CM_logand_context_with_test();
    }
    CM_u_move_zero_1L (temp,point_length*number);
    CM_load_context(context);
    CM_deallocate_stack_through(context);
  }
  */
  if (flags & PGL_print_sgi_inputs) sgi_flags |= 
    SGI_PRINT_INPUTS|SGI_PRINT_PROC_INFO;
  if (buffer->data_item_count == 0)
    buffer->data_items = (int*)malloc(sizeof(int));
  else 
    buffer->data_items =(int*)realloc(buffer->data_items,
				      (buffer->data_item_count+1)*sizeof(int));
  
  buffer->data_items[buffer->data_item_count]=SGI_write(number,sgi_flags,temp);
  buffer->data_item_count++;					       
  CM_deallocate_stack_through(temp);
}
