#include "pgl_internal.h"
#include <stdio.h>
  

PGL_spawn_frame_t PGL_open_spawn (dest,fill,format,length,size,flags)
     PGL_buffer_t dest;
     void (*fill)();
     PGL_format_t format;
     unsigned length,size,flags;
{
  PGL_spawn_frame_t frame=(PGL_spawn_frame_t)malloc(sizeof(PGL_spawn_frame));
  unsigned l_temp=length;
  CM_vp_set_id_t entry_vp_set=CM_current_vp_set;
  
  frame->flags = flags;
  frame->size = size;
  frame->length = length;
  frame->fill=fill;
  frame->format=format;
  frame->dest=dest;
  frame->pointer=frame->count=0;

  frame->geometry = CM_create_geometry (&l_temp,1);
  frame->vp_set =  CM_allocate_vp_set (frame->geometry);
  frame->seg= CM_allocate_heap_field_vp_set(1,frame->vp_set);
  frame->xh=CM_allocate_heap_field_vp_set(32,frame->vp_set);
  frame->xp=CM_allocate_heap_field_vp_set(frame->size,frame->vp_set);
  CM_set_vp_set (frame->vp_set);
  CM_u_move_zero_always_1L (frame->seg,1);
  CM_set_vp_set(entry_vp_set);
  return(frame);
}


void PGL_flush_spawn (frame)
     PGL_spawn_frame_t frame;
{
  CM_vp_set_id_t entry_vp_set=CM_current_vp_set;
  CM_field_id_t address;
  int i;

/*  printf ("click %d.\n",frame->pointer);*/
  if (frame->pointer > 0){
    CM_set_vp_set (frame->vp_set);
    CM_load_context (frame->seg);

    CM_set_context();
    if (frame->count){
      CM_spread_from_processor_1L (frame->xh,frame->length-1,frame->xh,32);
      CM_u_write_to_processor_1L (0,CM_context_flag,0,1);
    }
    
    if (frame->pointer < frame->length){
      address = CM_allocate_stack_field(32);
      CM_u_move_zero_always_1L (address,32);
      CM_my_send_address_1L(address,32);
      CM_u_lt_constant_1L (address,frame->pointer,32);
      CM_logand_context_with_test();
      CM_deallocate_stack_through(address);
    }

    CM_u_move_zero_1L (frame->xh,32);
    CM_logxor_constant_3_1L (frame->xh,frame->seg,1,1);
    if (frame->count)
      CM_u_write_to_processor_1L (0,CM_context_flag,1,1);
    _CMI_scan_with_copy_1d (frame->xp,frame->xp,CM_send_order,frame->size,
			    CM_upward,CM_inclusive,CM_start_bit,frame->seg);
    _CMI_scan_with_s_add_1d (frame->xh,frame->xh,CM_send_order,32,
			     CM_upward,CM_inclusive,
			     CM_start_bit,frame->seg);

    (*(frame->fill))(frame);
    CM_u_move_zero_always_1L(frame->seg,1);
    CM_set_vp_set (entry_vp_set);
    (frame->pointer)-=(frame->length);
    frame->count++;
  }
}


void PGL_write_spawn (frame,pixels,args)
     PGL_spawn_frame_t frame;
     CM_field_id_t pixels,args;
{
  CM_field_id_t entry_context = CM_allocate_stack_field (1);
  CM_field_id_t real_entry_context = CM_allocate_stack_field (1);
  CM_field_id_t address = CM_allocate_stack_field (32);
  CM_field_id_t other = CM_allocate_stack_field (32);
  CM_field_id_t nt = CM_allocate_stack_field(32);
  CM_field_id_t tseg;
  int total_procs_needed;
  int i,cont;
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  CM_geometry_id_t entry_geometry = CM_vp_set_geometry(entry_vp_set);

  CM_store_context (real_entry_context);
  if ((frame->flags)&PGL_clip_news_edges){
    for (i=0;i<CM_geometry_rank (entry_geometry);i++){
      CM_my_news_coordinate_1L (address,i,32);
      CM_u_lt_constant_1L (address,
			   CM_geometry_axis_length(entry_geometry,i)-1,32);
      CM_logand_context_with_test();
    }
  }

  CM_store_context (entry_context);  
  _CMI_scan_with_s_add_1d(other,pixels,CM_send_order,32,
			  CM_upward,CM_inclusive,CM_none,0);

  /* move this into a global_u_add in open spawn*/
  /*  printf ("total_procs:%d %d\n",CM_available_memory()*/
  total_procs_needed = CM_global_s_max_1L (other,32);
  
  if (total_procs_needed < 0) total_procs_needed=0;
  CM_s_subtract_3_1L (address,other,pixels,32);
  CM_s_add_constant_2_1L (address,frame->pointer,32);
  CM_s_add_constant_2_1L (other,frame->pointer,32);  
  frame->pointer += total_procs_needed;  

  tseg = CM_allocate_stack_field_vp_set(1,frame->vp_set);
  cont = 1;
  while (cont){
    CM_s_ge_constant_1L (other,0,32);
    CM_logand_context_with_test();
    CM_s_lt_constant_1L (address,frame->length,32);
    CM_logand_context_with_test();    
    CM_set_vp_set(frame->vp_set);
    CM_u_move_always_1L(tseg,frame->seg,1);
    CM_set_vp_set(entry_vp_set);
    
    CM_send_1L (frame->xp,address,args,frame->size,frame->seg);
    CM_set_vp_set(frame->vp_set);
    CM_logior_2_1L (frame->seg,tseg,1);
    CM_set_vp_set(entry_vp_set);

    if (frame->pointer >= frame->length)
      PGL_flush_spawn(frame);  
    else cont =0;

    CM_s_move_zero_1L (address,32);
    CM_invert_context();
    CM_logand_context (entry_context); 
    CM_s_subtract_constant_2_1L (address,frame->length,32);
    CM_load_context(entry_context);
    CM_s_subtract_constant_2_1L (other,frame->length,32);
  }
  CM_set_vp_set(frame->vp_set);
  CM_set_vp_set(entry_vp_set);
  CM_load_context(real_entry_context);
  CM_deallocate_stack_through (entry_context);
}

  
void PGL_close_spawn(frame)
     PGL_spawn_frame_t frame;
{
  PGL_flush_spawn(frame);
  CM_deallocate_heap_field(frame->seg);
  CM_deallocate_heap_field(frame->xh);
  CM_deallocate_heap_field(frame->xp);
  CM_deallocate_vp_set (frame->vp_set);
  CM_deallocate_geometry (frame->geometry);
  free(frame);
}

