#include "pgl_internal.h"
#include <stdio.h>

#define P_TYPE PGL_signed
#define P_LENGTH 32

#define SP S(P_LENGTH)
#define EP E(P_LENGTH)

extern void PGL_fill_line();

void PGL_fill_triangle(frame)
     PGL_spawn_frame_t frame;
{
  unsigned point_length = PGL_point_length(frame->format);
  unsigned xl = PGL_length (frame->format,PGL_x);
  unsigned yl = PGL_length (frame->format,PGL_y);
  unsigned xt = PGL_type (frame->format,PGL_x);
  unsigned yt = PGL_type (frame->format,PGL_y);

  CM_field_id_t x1=frame->xp;
  CM_field_id_t x2=CM_add_offset_to_field_id(frame->xp,point_length);
  CM_field_id_t x3=CM_add_offset_to_field_id(frame->xp,2*point_length);
  CM_field_id_t entry_context = CM_allocate_stack_field(1);
  CM_field_id_t total = CM_allocate_stack_field(P_LENGTH);
  CM_field_id_t ttemp = CM_allocate_stack_field(P_LENGTH);
  CM_field_id_t t = CM_allocate_stack_field(P_LENGTH);
  CM_field_id_t r1 = CM_allocate_stack_field(2*point_length);
  CM_field_id_t r2 = CM_add_offset_to_field_id(r1,point_length);

  PGL_cast (t,frame->xh,P_TYPE,PGL_signed,P_LENGTH,32);
  CM_store_context(entry_context);


  PGL_distance (total,PGL_extract(x1,frame->format,PGL_y),
		PGL_extract(x2,frame->format,PGL_y),
		P_TYPE,yt,P_LENGTH,yl);
  
  PGL_distance (ttemp,PGL_extract(x2,frame->format,PGL_y),
		PGL_extract(x3,frame->format,PGL_y),
		P_TYPE,yt,P_LENGTH,yl);
  
  PGL_gt (ttemp,total,P_TYPE,P_LENGTH);
  CM_logand_context_with_test();
  CM_u_move_1L (total,ttemp,P_LENGTH);
  CM_swap_2_1L (x1,x3,point_length);
  CM_load_context(entry_context);    
  
  PGL_distance(ttemp,PGL_extract(x1,frame->format,PGL_y),
	       PGL_extract(x3,frame->format,PGL_y),
		P_TYPE,yt,P_LENGTH,yl);
  PGL_gt (ttemp,total,P_TYPE,P_LENGTH);
  CM_logand_context_with_test();
  CM_u_move_1L (total,ttemp,P_LENGTH);
  CM_swap_2_1L (x2,x3,point_length);
  CM_load_context(entry_context);    

  PGL_gt (PGL_extract(x1,frame->format,PGL_y),
	  PGL_extract(x2,frame->format,PGL_y),yt,yl);
  CM_logand_context_with_test();
  CM_swap_2_1L (x1,x2,point_length);
  CM_load_context(entry_context);
  
  PGL_interpolate_field (r1,x1,x2,frame->format,t,total,P_TYPE,P_LENGTH);

  CM_load_context(entry_context);
  
  PGL_gt (PGL_extract(r1,frame->format,PGL_y),
	  PGL_extract(x3,frame->format,PGL_y),yt,yl);

  CM_logand_context_with_test();
  CM_swap_2_1L (x1,x2,point_length);
  
  switch(P_TYPE){
  case PGL_float:
    CM_f_subfrom_2_1L (t,total,S(P_LENGTH),E(P_LENGTH));
    CM_f_add_constant_2_1L (t,1.0,S(P_LENGTH),E(P_LENGTH));
    break;
  case PGL_signed:
    CM_s_subfrom_2_1L (t,total,P_LENGTH);
    CM_s_add_constant_2_1L (t,1,P_LENGTH);
    break;
  case PGL_unsigned:
    CM_u_subfrom_2_1L (t,total,P_LENGTH);
    CM_u_add_constant_2_1L (t,1,P_LENGTH);
    break;
  }

  CM_load_context(entry_context);
  /* this is extraneous*/

  PGL_distance(ttemp,PGL_extract(x1,frame->format,PGL_y),
	       PGL_extract(x3,frame->format,PGL_y),
	       P_TYPE,yt,P_LENGTH,yl);

  PGL_interpolate_field (r2,x1,x3,frame->format,t,ttemp,P_TYPE,P_LENGTH);

  if (frame->flags & PGL_use_scan_lines)
    PGL_write_scan_line (frame->next,r1);
  else{
    PGL_distance (total,PGL_extract(r1,frame->format,PGL_x),
		  PGL_extract(r2,frame->format,PGL_x),
		  PGL_signed,xt,32,xl);
    PGL_write_spawn (frame->next,total,r1);
  }
  CM_deallocate_stack_through (entry_context);
}


void PGL_plot_triangle_internal (frame,dest,format,source)
     PGL_spawn_frame_t frame;
     PGL_buffer_t dest;
     PGL_format_t format;
     CM_field_id_t source;
{
  unsigned point_length= PGL_point_length(format);
  CM_field_id_t entry_context = CM_allocate_stack_field(1);
  CM_field_id_t pixels = CM_allocate_stack_field(32);
  CM_field_id_t temp = CM_allocate_stack_field(32);
  CM_field_id_t x1=source;
  CM_field_id_t x2=CM_add_offset_to_field_id(source,point_length);
  CM_field_id_t x3=CM_add_offset_to_field_id(source,2*point_length);
  unsigned yl = PGL_length(format,PGL_y);
  unsigned yt = PGL_type(format,PGL_y);


  CM_store_context(entry_context);
  CM_store_context(pixels);
  PGL_clip_line(pixels,dest,format,x1,x2);
  CM_store_context(temp);
  PGL_clip_line(temp,dest,format,x2,x3);
  CM_logior_2_1L (pixels,temp,1);
  CM_store_context(temp);
  PGL_clip_line(temp,dest,format,x3,x1);
  CM_logior_2_1L (pixels,temp,1);
  CM_load_context(pixels);

  PGL_distance (pixels,PGL_extract(x1,format,PGL_y),
		PGL_extract(x3,format,PGL_y),
		PGL_signed,yt,32,yl);
  
  PGL_distance (temp,PGL_extract(x2,format,PGL_y),
		PGL_extract(x3,format,PGL_y),
		PGL_signed,yt,32,yl);

  CM_s_max_2_1L (pixels,temp,32);

  PGL_distance (temp,PGL_extract(x2,format,PGL_y),
		PGL_extract(x1,format,PGL_y),
		PGL_signed,yt,32,yl);
  
  CM_s_max_2_1L (pixels,temp,32);
  PGL_write_spawn (frame,pixels,source);
  CM_load_context(entry_context);
  CM_deallocate_stack_through(entry_context);
}


PGL_spawn_frame_t PGL_open_triangle (dest,format,size1,size2,flags)
     PGL_buffer_t dest;
     PGL_format_t format;
     unsigned size1,size2;
     unsigned flags;
{
  PGL_format_t i_f=PGL_generate_transfer_format(dest,format,flags);
  unsigned point_length = PGL_point_length(i_f);
  PGL_spawn_frame_t frame;
  
  frame=PGL_open_spawn(dest,PGL_fill_triangle,i_f,size1,3*point_length,flags);
  
  if (frame->flags & PGL_use_scan_lines)
    frame->next=PGL_open_scan_line(dest,i_f,2*point_length,flags);
  else
    frame->next=PGL_open_spawn(dest,PGL_fill_line,i_f,size2,
			       2*point_length,flags);

  return (frame);
}


void PGL_write_triangle (frame,format,p1,p2,p3)
     PGL_spawn_frame_t frame;
     PGL_format_t format;
     CM_field_id_t p1,p2,p3;
{
  unsigned point_length = PGL_point_length (frame->format);
  CM_field_id_t x1=CM_allocate_stack_field (3*point_length);
  CM_field_id_t x2=CM_add_offset_to_field_id(x1,point_length);
  CM_field_id_t x3=CM_add_offset_to_field_id(x2,point_length);

  PGL_generate_transfer_field(x1,p1,frame->dest,frame->format,format,
			      frame->flags,0);
  PGL_generate_transfer_field(x2,p2,frame->dest,frame->format,format,
			      frame->flags,0);
  PGL_generate_transfer_field(x3,p3,frame->dest,frame->format,format,
			      frame->flags,0);

  PGL_plot_triangle_internal (frame,frame->dest,frame->format,x1);
  CM_deallocate_stack_through (x1);
}


void PGL_close_triangle (frame)
     PGL_spawn_frame_t frame;
{
  PGL_flush_spawn (frame);
  if (frame->flags & PGL_use_scan_lines)
    PGL_close_scan_line (frame->next);
  else
    PGL_close_spawn(frame->next);
  PGL_close_spawn(frame);
}


void PGL_plot_triangle(dest,flags,format_string,p1,p2,p3)
     PGL_buffer_t dest;
     unsigned flags;
     char *format_string;
     CM_field_id_t p1,p2,p3;
{
  PGL_format_t format=(PGL_format_t)PGL_translate_format(format_string);
  PGL_spawn_frame_t frame=PGL_open_triangle(dest,format,262144,262144,flags);
  PGL_write_triangle (frame,format,p1,p2,p3);
  PGL_close_triangle(frame);
  PGL_free_format (format);
}

