/*   File Header   ---------------------------------
     File Name: scene.c
     Author and Date: Eric Hoffman 2/22/90
     Description: main routine for scene generation
    End Header   ---------------------------------- */

#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <math.h>
#include <pgl/pgl.h>
#include <pgl/pgl_internal.h>

static unsigned block_dim[32];
static rendering_flags=PGL_buffer_normalization|PGL_print_sgi_inputs;

#define STRETCH 1


void GS_allocate_scene_buffer(scene)
     GS_scene_t scene;
{
  char format[40];
  float xsize;
  float ysize;
  
#ifdef STRETCH
  xsize=(float)scene->dimensions[0]/2.0;
  ysize=(float)scene->dimensions[1]/2.0;
#else
  xsize=ysize = (float)MIN(scene->dimensions[0],scene->dimensions[1])/2.0;
#endif

  sprintf (format,"ru8 gu8 bu8 zu%d Tu%d",scene->depth_bits,
	   scene->priority_bits);
  scene->buffer=(char *)PGL_allocate_buffer(scene->dimensions[0],
					    scene->dimensions[1],
					    (1<<scene->depth_bits),
					    (scene->remote_render_t)?
					    PGL_sgi_render:PGL_max,
					    format);
					    
  PGL_buffer_ambient (scene->buffer,
		      (int)(255.0*scene->ambient),
		      (int)(255.0*scene->ambient),
		      (int)(255.0*scene->ambient));
  PGL_clear_buffer (scene->buffer,
		    scene->background[0],
		    scene->background[1],
		    scene->background[2]);
  if (scene->light_type == GS_POINT_LIGHT){
    PGL_add_light (scene->buffer,
		   scene->light_source[0],
		   scene->light_source[1],
		   scene->light_source[2],255,255,255);
    PGL_add_light (scene->buffer,
		   -scene->light_source[0],
		   -scene->light_source[1],
		   -scene->light_source[2],255,255,255);
  }

  PGL_translate_buffer(scene->buffer,PGL_x,-scene->center[0]);
  PGL_translate_buffer(scene->buffer,PGL_y,-scene->center[1]);
  PGL_translate_buffer(scene->buffer,PGL_z,-scene->center[2]);

  PGL_scale_buffer (scene->buffer,PGL_x,1.0/pow(GS_ZOOM_RATIO,scene->zoom));
  PGL_scale_buffer (scene->buffer,PGL_y,1.0/pow(GS_ZOOM_RATIO,scene->zoom));

  PGL_translate_buffer(scene->buffer,PGL_z,1.0);
  PGL_scale_buffer(scene->buffer,PGL_z,(float)((1<<(scene->depth_bits-1))-4));
  PGL_scale_buffer(scene->buffer,PGL_x,xsize);
  PGL_scale_buffer(scene->buffer,PGL_y,-ysize);
  PGL_translate_buffer(scene->buffer,PGL_x,xsize);
  PGL_translate_buffer(scene->buffer,PGL_y,ysize);
}


void GS_update_scene(scene_id)
     GB_id scene_id;
{  
  GS_scene_t scene = (GS_scene_t)GB_get_object(scene_id);
  GS_rep_list_t rep_pos;
  GS_representation_t rep;
  GD_data_desc *desc;
  CM_field_id_t data;
  CM_vp_set_id_t entry_vp_set = CM_current_vp_set;
  unsigned flags;

  if (scene-> time_this_scene){
    CM_start_timer(0);
    CM_reset_timer(0);
  }
  flags=rendering_flags;
  if (scene->buffer != 0) PGL_deallocate_buffer(scene->buffer);
  GS_allocate_scene_buffer(scene);
  for(rep_pos=scene->representations;rep_pos!=0;rep_pos=rep_pos->next){
    rep = (GS_representation_t)GB_get_object(rep_pos->rep_id);
    desc = (GD_data_desc *)GB_get_object(rep_pos->data_id);
    GD_open_ds(desc,scene->projection,scene->parameters,block_dim);
    GS_open_rep(rep->rep_code,scene->buffer,rep_pos->priority,block_dim,
		rep_pos->rendering_memory,flags);
    while ((data=GD_read_ds(desc,scene->projection,scene->parameters,
			    block_dim))!=CM_no_field){
      GS_write_rep(rep->rep_code,data,rep,block_dim);
    }
    GS_close_rep(rep->rep_code);
    GD_close_ds (desc);
  } 
  /*  CM_set_vp_set(((PGL_buffer_t)scene->buffer)->vp_set);
      PGL_print_field(((PGL_buffer_t)scene->buffer)->field,
      ((PGL_buffer_t)scene->buffer)->format);*/
  CM_set_vp_set(entry_vp_set);
  if (scene->time_this_scene) CM_stop_timer(1);
}

