#include "xtool.h"

extern GB_entry GB_hash_table[];
extern Widget widget_array[];
extern GX_widget_list_t movies;
extern GX_scene_list_t scenes;
extern GX_scene_list_t curr_scene;
extern GX_xtool_projection_t curr_proj;
extern int cont_reset;
extern int file_based;
extern int cont_updt;
extern int cont_disp;
extern char RDformat[];
extern char curr_display[];
extern int x_offset;
extern int y_offset;
extern FILE *x_out;

extern void movie_callback();
extern Widget XmCreatePushButton();

/**************************************

 GX_string_create
 GX_string_read
 GX_clear_list
 GX_add_list_item
 GX_create_object
 GX_reset_widget_list
 GX_clear_movie_list
 GX_add_movie_button
 GX_reset_movie_list
 GX_reset_controls
 GX_reset_new_scene

  GX_Update_Scene 
    builds a glisp string necessary to update a scene.
  GX_Display_Scene 
    builds a glisp string necessary to display a scene.
  GX_Output_Command
    Prints the command to the output file*.
    Also FREEs the command structure.
  GX_get_instance
    Returns the specified instance from the GB data base.
  GX_get_label
    Takes a GB_id and place to put a label, and fills in the label.
  GX_reset_object
    Instructs the host application to print an object, and resets the
    object in the local DB.
  GX_reset_data
    Instructs the host application to print a list of all the objects
    and builds them if necessary in the local DB.
  GX_init_data
    Initializes DB with GB_init, and resets the data.
  GX_list_objects
    returns a list of all labels in DB, or all labels of a particular type.

  create_callback

 **************************************/

int power (base, exp)
     int base, exp;
{ int i,result = 1;
  for (i=0; i<exp; i++) result *= base;
  return (result);
}

void GX_update_scene (com)
     char com[];
{
  sprintf (com, "(update_scene X_UP_SC (scene_id %s))", curr_scene->label);
  if ((cont_disp) || (cont_updt)) {
    sprintf (com, "%s\n\n(display_scene X_DIS_SC\n\t(scene_id %s)\n",
	     com, curr_scene->label);
    sprintf (com, "%s\t(display_id %s)\n\t(offset %d %d))",
	     com, curr_display, curr_scene->x_off, curr_scene->y_off);
  }
}

void GX_display_scene (com)
     char com[];
{
  sprintf (com, "(display_scene X_DIS_SC\n\t(scene_id %s)\n",
	   curr_scene->label);
  sprintf (com, "%s\t(display_id %s)\n\t(offset %d %d))",
	   com, curr_display, curr_scene->x_off, curr_scene->y_off);
}

void GX_output_command (command, more, comment)
     char command[];
     int more, comment;
{
  fprintf (x_out, command);
  if (!more) {
    fprintf (x_out, "\n");
    if ((cont_updt) && (!comment)) {
      sprintf (command, "");
      GX_update_scene (command);
      fprintf (x_out, command);
      fprintf (x_out, "\n\n");
    }
    fflush (x_out);
    free (command);
  }
  fprintf (x_out, "\n");
  fflush (x_out);
}

XmString GX_string_create (text)
     char *text;
{
  return (XmStringCreate (text, XmSTRING_DEFAULT_CHARSET));
}

Boolean GX_string_read (xstring, text)
     XmString xstring;
     char *text[];
{
  short parsed = 0;
  sprintf (text, "");
  parsed = XmStringGetLtoR (xstring, XmSTRING_DEFAULT_CHARSET, text);
  if (!parsed)
    sprintf (text, "");
  return (parsed);
}

int *GX_get_instance (place,type,label)
     char place[];
     unsigned type;
     char label[];
{
  char etext[GX_COM_LEN];
  GB_entry *entry;
  GB_id id; 
  int *result, error = 1;

  result = NULL;
  id = GB_get_id (label, type);
  if (id == 0)
    sprintf (etext, ";;; Error in %s, label unknown %s",place,label);
  else {
    entry = (GB_entry *)GB_get_entry(id);
    if (entry == NULL)
      sprintf (etext, ";;; Error in %s, ID unknown %d",place,id);
    else {
      if(entry->type != type)
	sprintf(etext,";;; Error in %s ID %d, %s, is type %d: expected %d",
		place, id, label, entry->type, type);
      else {
	error = 0;
	result = entry->instance;
      }
    }
  }
  if (error)
    GX_output_command (etext,0,1);
  return(result);
}

void GX_get_label (place, id, text)
     char place[];
     GB_id id;
     char *text;
{
  GB_entry *entry;

  entry = (GB_entry *)GB_get_entry(id);
  if (entry == NULL) {
    char etext[GX_COM_LEN];
    sprintf (etext, ";;; Error in %s, ID unknown %d",place,id);
    GX_output_command (etext,0,1);
  }
  else {
    sprintf (text, "%s", entry->label);
  }
}

void GX_clear_list (list)
     Widget list;
{
  Arg args[1];
  int i, count = 0;
  XtSetArg (args[0], XmNitemCount, &count);
  XtGetValues (list, args, 1);
  for (i=0; i<count; i++)
    XmListDeletePos (list, 0);
}

void GX_add_list_item (list, label)
     Widget list;
     char *label;
{
  XmString string;
  string = GX_string_create (label);
  XmListAddItem (list, string, 0);
}

void GX_create_object (label, type)
     char* label;
     unsigned type;
{
  Widget list;
  char command[GX_COM_LEN];
  if (!(GX_get_instance ("GX_create_object", type, label))) {
    switch (type) {
    case GB_SCENE:
      sprintf (command, "(scene %s)", label);
      list = widget_array[k_se_sclist];
      break;
    case GB_REPRESENTATION:
      sprintf (command, "(representation %s)", label);
      list = widget_array[k_se_replist];
      break;
    case GB_DATA_DESC:
      sprintf (command, "(data_desc %s)", label);
      list = widget_array[k_se_datlist];
      break;
    }
    GB_add_object (label, type);
    GX_add_list_item (list, label);
    GX_output_command (command,0,0);
  }
}   

void GX_reset_object (label)
     char* label;
{
  if ((!file_based) && (cont_reset)) {
    char command[GX_COM_LEN];
    sprintf (command, "(print X_OBJECT (id %s))", label);
    GX_output_command (command,0,0);
/*    GB_parse_file (x_in, x_out, x_error);*/
  }
}

void GX_reset_data ()
{
  if ((!file_based) && (cont_reset)) {
    /* Duplicate the data that is in the calling application's DB */
    char command[GX_COM_LEN];
    sprintf (command, "(print_objects X_OBJECTS)");
    GX_output_command (command,0,0);
  }
  else {
    /* Load in same init files as DB does */
    char pathname[128];
    sprintf (pathname,"%s/line_styles.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
    sprintf (pathname,"%s/symbols.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
    sprintf (pathname,"%s/data_sets.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
    sprintf (pathname,"%s/representations.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
    sprintf (pathname,"%s/scenes.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
    sprintf (pathname,"%s/video.gl",GEODATA_LIB);
    GB_load_object_file(pathname);
  }
}

void GX_init_data ()
{
  char pathname[128];

  GB_init();
  GX_reset_data();

  /* Load in Xtool initialization glisp file */
  sprintf (pathname,"%s/xtool.gl",GEODATA_LIB);
  GB_load_object_file(pathname);
}

GX_label_list *GX_list_objects (type)
     unsigned type;
{
  GX_label_list *item, *labels = NULL;
  GB_entry *j;
  int i,k;
  char name[GB_LABEL_LENGTH];
  
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    for (j = GB_hash_table+i; j != (GB_entry *)0; j=j->next)
      if ((j->id) && ((type == 0) || (j->type == type))) {
	item = (GX_label_list *) malloc (sizeof (GX_label_list));
	item->next = labels;
	labels = item;
	sprintf (item->label, "%s", j->label);
	item->object_id = j->id;
      }
  return (labels);
}

void GX_reset_widget_list (list, type)
     Widget list;
     unsigned type;
{
  GX_label_list *item, *labels = NULL;

  GX_clear_list (list);
  labels = GX_list_objects (type);
  while (labels != NULL) {
    item = labels;
    labels = labels->next;
    GX_add_list_item (list, item->label);
    free (item);
  }
}

void GX_clear_movie_list ()
{
  GX_widget_list_t mov;
  while (movies != NULL) {
    mov = movies;
    movies = movies->next;
    XtDestroyWidget (mov->widget);
    free (mov);
  }
}

void GX_add_movie_button (label, id)
     char *label;
     GB_id id;
{
  Widget button;
  GX_widget_list_t movie;
  Arg args[5];
  XmString string;
  char pbname[GB_LABEL_LENGTH+3];

  sprintf (pbname, "%s_pb", label);
  string = GX_string_create (label);
  XtSetArg (args[0], XmNlabelString, string);
  button = XmCreatePushButton (widget_array[k_m_movie], pbname, args, 1);
  XtAddCallback (button, XmNactivateCallback, movie_callback, (caddr_t)0);
  movie = (GX_widget_list *) malloc (sizeof (GX_widget_list));
  movie->widget = button;
  movie->next = movies;
  movies = movie;
  XtManageChild (button);
}

void GX_reset_movie_list ()
{
  GX_label_list *item, *labels = NULL;

  GX_clear_movie_list ();
  labels = GX_list_objects (GB_MOVIE);
  while (labels != NULL) {
    item = labels;
    labels = labels->next;
    GX_add_movie_button (item->label, item->object_id);
    free (item);
  }
}

void GX_reset_controls ()
{
  char rep[GB_LABEL_LENGTH], dat[GB_LABEL_LENGTH];
  char text[3*GB_LABEL_LENGTH], tag[GB_LABEL_LENGTH];
  int i, j, n, d;
  Widget rdlist;
  Arg args[10];
  XmString str;
  GS_scene *scene = NULL;
  GS_rep_list *rep_list = NULL;
  GB_id id;

  scene = (GS_scene *)
    GX_get_instance ("GX_reset_controls", GB_SCENE, curr_scene->label);

  /* Reset all xtool controls based on a new scene */
  /* Set Rep / Dat list */
  if (scene != NULL) {
    rep_list = scene->representations;
    rdlist = widget_array[k_se_rdlist];
    GX_clear_list (rdlist);
    while (rep_list != NULL) {
      id = rep_list->rep_id;
      GX_get_label ("GX_reset_controls", id, rep);
      id = rep_list->data_id;
      GX_get_label ("GX_reset_controls", id, dat);
      sprintf (text, RDformat, rep, dat);
      GX_add_list_item (rdlist, text);
      rep_list = rep_list->next;
    }

    /* Set Display X and Y offset scales and text fields in Control Panel */
    sprintf (tag, "X Off: %d", curr_scene->x_off);
    XtSetArg (args[0], XmNvalue, tag);
    XtSetValues(widget_array[k_cp_xoff_tx], args, 1);
    XtSetArg (args[0], XmNvalue, ((curr_scene->x_off) / 16));
    XtSetValues(widget_array[k_cp_xoff_sc], args, 1);
    sprintf (tag, "Y Off: %d", curr_scene->y_off);
    XtSetArg (args[0], XmNvalue, tag);
    XtSetValues(widget_array[k_cp_yoff_tx], args, 1);
    XtSetArg (args[0], XmNvalue, ((curr_scene->y_off) / 16));
    XtSetValues(widget_array[k_cp_yoff_sc], args, 1);

    /* Set Scene size and zoom scales and text fields in Scene Editor */
    sprintf (tag, "Scene Width: %d", scene->dimensions[0]);
    XtSetArg (args[0], XmNvalue, tag);
    XtSetValues(widget_array[k_se_width_tx], args, 1);
    XtSetArg (args[0], XmNvalue, (scene->dimensions[0] / 16));
    XtSetValues(widget_array[k_se_width_sc], args, 1);
    sprintf (tag, "Scene Height: %d", scene->dimensions[1]);
    XtSetArg (args[0], XmNvalue, tag);
    XtSetValues(widget_array[k_se_height_tx], args, 1);
    XtSetArg (args[0], XmNvalue, (scene->dimensions[1] / 16));
    XtSetValues(widget_array[k_se_height_sc], args, 1);
    sprintf (tag, "Scene Zoom: %g", scene->zoom);
    XtSetArg (args[0], XmNvalue, tag);
    XtSetValues(widget_array[k_se_zoom_tx], args, 1);
    XtSetArg (args[0], XmNvalue, (int)((scene->zoom)*10));
    XtSetValues(widget_array[k_se_zoom_sc], args, 1);

    /* Set Projection Radio Button */
    switch (scene->projection) {
    case GS_NO_PROJECTION: sprintf (tag, "NO_PROJECTION"); break;
    case GS_MILLER: sprintf (tag, "MILLER"); break;
    case GS_MOLLWEIDE: sprintf (tag, "MOLLWEIDE"); break;
    case GS_LAMBERT: sprintf (tag, "LAMBERT"); break;
    case GS_POLAR_STEREO: sprintf (tag, "POLAR_STEREO"); break;
    case GS_GNOMONIC: sprintf (tag, "GNOMONIC"); break;
    case GS_GLOBE_VIEW: sprintf (tag, "GLOBE_VIEW"); break;
    case GS_GENERAL_PERSPECTIVE: sprintf (tag, "GENERAL_PERSPECTIVE"); break;
    }
    if (curr_proj) {
      XtSetArg (args[0], XmNset, FALSE);
      XtSetValues
	(widget_array[(curr_proj->GX_proj_code) - GS_NO_PROJECTION +
		      k_pp_mill_tb - 1], args, 1);
    }
    curr_proj = (GX_xtool_projection_t)
      GX_get_instance ("projection_callback", GB_XTOOL_PROJECTION, tag);
    XtSetArg (args[0], XmNset, TRUE);
    XtSetValues (widget_array[(curr_proj->GX_proj_code) - GS_NO_PROJECTION +
			      k_pp_mill_tb - 1], args, 1);

    /* Set projection parameter scales */
    for (j=0; j<NO_PROJECTION_PARAMS; j++) {
      for (i=0; i<GX_NO_SCALES; i++)
	if (curr_proj->GX_scale_map[i] == j) {
	  curr_proj->GX_defaults[j] = ((scene->parameters[j]) /
				       (curr_proj->GX_convert[j]));
	  sprintf (tag, "%s: %.*f", curr_proj->GX_labels[j],
		   (curr_proj->GX_decimals[j]),
		   (curr_proj->GX_defaults[j]));
	  XtSetArg (args[0], XmNvalue, tag);
	  XtSetValues(widget_array[i+k_pp_par1_tx], args, 1);
	  n = 0;
	  d = power (10, (curr_proj->GX_decimals[j]));
	  XtSetArg (args[n], XmNminimum,(int)(curr_proj->GX_min[j]*d)); n++;
	  XtSetArg (args[n], XmNmaximum,(int)(curr_proj->GX_max[j]*d)); n++;
	  XtSetArg (args[n], XmNvalue,(int)(curr_proj->GX_defaults[j]*d)); n++;
	  XtSetArg (args[n], XmNscaleMultiple,
		    (int)(((curr_proj->GX_max[j]-curr_proj->GX_min[j])*d)/20));
	  n++;
	  XtSetValues(widget_array[i+k_pp_par1_sc], args, n);
	}
    }
  }
}

void GX_reset_new_scene ()
{
  char rep[GB_LABEL_LENGTH], dat[GB_LABEL_LENGTH];
  char text[3*GB_LABEL_LENGTH];
  int j, width, height, zoom;
  Widget rdlist;
  Arg args[10];
  GS_scene *scene = NULL;
  GS_rep_list *rep_list = NULL;
  GB_id id;
  char command[GX_COM_LEN];

  scene = (GS_scene *)
    GX_get_instance ("GX_reset_new_scene", GB_SCENE, curr_scene->label);

  /* Change scene to match xtool control values */
  /* Don't change scene's Set Rep / Dat list */
  if (scene) {
    rep_list = scene->representations;
    rdlist = widget_array[k_se_rdlist];
    GX_clear_list (rdlist);
    while (rep_list != NULL) {
      id = rep_list->rep_id;
      GX_get_label ("GX_reset_new_scene", id, rep);
      id = rep_list->data_id;
      GX_get_label ("GX_reset_new_scene", id, dat);
      sprintf (text, RDformat, rep, dat);
      GX_add_list_item (rdlist, text);
      rep_list = rep_list->next;
    }

    /* Set GX internal scene with current X and Y offsets. */
    curr_scene->x_off = x_offset;
    curr_scene->y_off = y_offset;

    /* Set Scene size */
    XtSetArg (args[0], XmNvalue, &width);
    XtGetValues(widget_array[k_se_width_sc], args, 1);
    XtSetArg (args[0], XmNvalue, &height);
    XtGetValues(widget_array[k_se_height_sc], args, 1);
    if (scene) {
      scene->dimensions[0] = width * 16;
      scene->dimensions[1] = height * 16;
    }
    sprintf (command, "(scene %s (dimensions %d %d))",
	     curr_scene->label, (width * 16), (height * 16));
    GX_output_command (command,1,0);

    /* Set Scene zoom */
    if (scene) {
      XtSetArg (args[0], XmNvalue, &zoom);
      XtGetValues(widget_array[k_se_zoom_sc], args, 1);
      scene->zoom = zoom / 10.0;
      sprintf (command, "(scene %s (zoom %d))",
	       curr_scene->label, scene->zoom);
      GX_output_command (command,1,0);
    }

    /* Set Scene Projection and Parameters*/
    if (curr_proj && scene) {
      scene->projection = curr_proj->GX_proj_code;
      sprintf (command, "(scene %s\n\t(projection %d)\n\t(parameters",
	       curr_scene->label, curr_proj->GX_proj_code);
      for (j=0; j<NO_PROJECTION_PARAMS; j++) {
	scene->parameters[j] = ((curr_proj->GX_defaults[j]) *
				(curr_proj->GX_convert[j]));
	sprintf (command, "%s %f", command, scene->parameters[j]);
      }
      sprintf (command, "%s))", command);
      GX_output_command (command,0,0);
    }
  }
}

void create_callback(widget, tag, data)
     Widget widget;
     int *tag;
     XmAnyCallbackStruct *data;
{
  widget_array[*tag] = widget;
  switch (*tag) {
  case k_se_sclist: GX_reset_widget_list (widget, GB_SCENE); break;
  case k_se_replist: GX_reset_widget_list (widget, GB_REPRESENTATION); break;
  case k_se_datlist: GX_reset_widget_list (widget, GB_DATA_DESC); break;
  }
}
